# SCRIPT FOR SENDING PUSHOVER NOTIFICATIONS WITH  WITH A GIF FILE ATTACHMENT
#
# Version: 1.9
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V1.9 - 2023-09-07 - FIXED BUG: SendPushover() call lacked grave accent (`) after -Camera param line; caused malformed value
#                   - FIXED BUG: all DAT-extracted JPG files lacked JPEG Interchange Format end-of-image byte marker ('FF D9')
#                   - IMPROVED: JPG validation function 'IsJpegFileValid()' - now also tests for valid end-of-image byte marker
#                   - ADDED: repair attempted for situations where the JPG start- & end-of-image byte markers appear to be unbalanced
#                   - ADDED" -Help argument (usage guide no longer displayed by default)
#                   - DISABLED: argument -NukeDatFile has been disabled until Blue Iris provides a safe method to delete DAT files
# V1.8 - 2023-07-12 - fixed bug; -Priority parameter (mis-typed as [string] instead of [int]; PS7 forgiving; PS5 not)
#                   - improved handling of warning if user appears to be trying to use parameter -ImagesCnt with DAT files
# V1.7 - 2023-07-09 - fixed bug; BI logfile message (missing source file)
#                   - added error trapping for non-existent database record IDs
#                   - shortened message in BI logfile, and 'gif_source' entry in CSV log (removed path)
#                   - changed logged GIF duration values units from msec to seconds (REQUIRES RENAMING OLD CSV FILE)
# V1.6 - 2023-06-30 - improved error trapping & feedback for DAT files with insufficient images
#                   - improved error trapping when trying to get $gif_jpg_WxH from one of the JPG files
#                   - fixed handling of function ExtractDatImages() when its returned object is empty
#                   - added warning if user appears to be trying to use parameter -ImagesCnt with DAT files
#                   - improved how ffmpeg issues (if any) are displayed in script output to the console
# V1.5 - 2023-06-28 - fixed bug; function CopyFileAsSequence()
#                   - performance log (*.csv file) now reporting GIF's JPG WxH instead of Alert's WxH
#                   - function IsJpegFileValid() is now releasing resources ($img.Dispose()) to reduce risk of memory leak
# V1.4 - 2023-06-27 - fixed GIF size reduction algorithm to meet the Pushover API limit (GIF can no longer have <1 image)
#                   - added error trap; when ffmpeg fails to successfully create a GIF file
#                   - Blue Iris logfile logging levels now reflect the script's output if it contains 'ERR/' or 'WARNING:'
# V1.3 - 2023-06-27 - added switch -NoArchiveDat
#                   - fixed bug: prevented logging the source file & alert record to the CSV performance log file
# V1.2 - 2023-06-26 - added sleep status bar SleepWithStatusBar()
#                   - fixed bug; problem retrieving actively recording BVR filename (PowerShell v5)
# V1.1 - 2023-06-26 - changed default user setting '$gif_trim_end_custom'
# V1   - 2023-06-26 - initial release
#
# REQUIRES:
#    CHOCOLATEY - Download and install the Chocolatey package manager: https://chocolatey.org/install 
#    FFMPEG     - Install the latest ffmpeg build using [choco install ffmpeg] ... https://www.gyan.dev/ffmpeg/builds/
#               - Home: https://www.ffmpeg.org/download.html          
#
# USAGE:      {this script name}.ps1 Msg [Priority AttachFile Quality Width, TargetKB Quiet]
# ARGUMENTS:  -Msg           = [string] Pushover notification message - may contain HTML
#             -AlertRec      = (string) alert record, typ. via Blue Iris macro &ATTACH_DB
#             -AlertDB       = (string) alert record, typ. via Blue Iris macro &ATTACH_DB
#             -AlertMemo     = (string) alert AI memo, typ. via Blue Iris macro &MEMO
#             -Live          = (switch) create GIF from specified camera's stream
#             -Camera        = (string) camera shortname (required with -Live argument)
#             -UI3Rec        = (string) hyphenated numeric values following UI3 URL argument '&rec=', like 'ui3.htm?rec=208953322-1181690'
#             -DbRec         = (string) Blue Iris database record for clip, like 208953322
#             -Seconds       = [integer] GIF duration; overrides default - alert's duration, or 20 sec with -Live argument
#             -Priority      = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
#             -Title         = (string) custom title for pushover message
#             -IgnoreDatFile = (switch) ignores 'DAT' file (if it exists); 
#                                       by default, a DAT file is always used if it contains N or more JPGs
#                                       N is specified in the user-settings file '$gif_DAT_reqd_images'; the default is 4
#             -NukeDatFile   = (switch) removes DAT file from Blue Iris storage (always archived first to '\_dat_archive\') 
#                                       NOTE: console also puts file in the Windows Recycle Bin
#             -NoArchiveDat  = (switch) disables automatic archiving of DAT files (if used to create GIF) (enabled by default)
#             -DelayPct      = (integer) percentage of alert's duration to wait before extracting JPGs from clip
#             -ImagesCnt     = (integer) count of JPGs to use in the GIF animation
#             -Height        = (integer) GIF'S height dimension; units pixels
#             -FPS           = (integer) GIF frames per second
#             -TrimBegPct    = (integer) percentage of alert clip to trim from GIF animation (left side)
#             -TrimEndPct    = (integer) percentage of alert clip to trim from GIF animation (right side)
#             -TrimEndCust   = (object) applied only when alert 'memo' contains item in user-setting 
#             -NoPush        = (switch) disables sending Pushover notification (enabled by default)
#                                       NOTE: also disables management of GIF size to under 2.5 MB
#             -NoBiLog       = (switch) disables logging to the blue Iris logfile (enabled by default)
#             -NoArchiveGif  = (switch) disables automatic archiving of GIF files (enabled by default)
#             -Quiet         = (switch) disables ALL background messaging, and all Console messaging EXCEPT warnings and return object
#             -Demo          = (switch) disables sending notifications, and posting to the Blue logfile
#             -Help          = (switch) enable display of usage guide
#             -RtnHashtable  = (switch) returns Powershell hashtable instead of default JSON string
# RETURNS:    (hashtable) @{}
#
# EXAMPLES:
#    To send Pushover with GIF from Blue Iris 'On alert...' action
#       Parameters: "C:\ps_scripts\BI_gof_tool.ps1" -AlertRec `&ALERT_DB -Msg """<b>&CAM</b> &MEMO <a href='&WAN/alerts/&ALERT_DB?fulljpeg'>Hi-Res</a> <a href='&WAN/ui3.htm?rec=&ALERT_DB'>UI3</a>""" -Priority 2 -UseDatFile -ImagesCnt 10 -FPS 2 -Height 320 -TrimBegPct 0 -TrimEndPct 50
#

# MUST BE THE FIRST LINE IN A SCRIPT
param(
	[string]$Msg,
	[string]$AlertRec,
	[string]$AlertDB,
	[string]$AlertMemo,
	[switch]$Live,
	[string]$Camera,
	[int]$Seconds,    # Default = 20
	[string]$UI3Rec,
	[string]$DbRec,
	[int]$Priority,
	[string]$Title,
	[switch]$IgnoreDatFile,
	[switch]$NukeDatFile,
	[switch]$NoArchiveDat,
	[int]$DelayPct,   # Default = 100
	[int]$ImagesCnt,  # Default = 10
	[int]$Height,     # Default = 320
	[int]$FPS,        # Default = 2
	[int]$TrimBegPct,
	[int]$TrimEndPct,#
	[int]$TrimEndCust,
	[switch]$NoPush,
	[switch]$NoBiLog,
	[switch]$NoArchive,
	[switch]$Quiet,
	[switch]$Demo,
	[switch]$Help,
	[switch]$RtnHashtable
)

# For debugging
#Write-Host ("PSBoundParameters = " + ($PSBoundParameters | ConvertTo-Json -Depth 100))
#Write-Host "Msg = $Msg"
#Write-Host "AlertRec = $AlertRec"
#Write-Host "AlertDB = $AlertRec"
#Write-Host "AlertMemo = $AlertMemo"
#Write-Host "Live = $Live"
#Write-Host "Camera = $Camera"
#Write-Host "DbRec = $UI3Rec"
#Write-Host "UI3Rec = $UI3Rec"
#Write-Host "Seconds = $Seconds"
#Write-Host ("Seconds = " + $Seconds.GetType())
#Write-Host "Priority = $Priority"
#Write-Host "Title = $Title"
#Write-Host "IgnoreDatFile = $IgnoreDatFile"
#Write-Host "NukeDatFile = $NukeDatFile"
#Write-Host "DelayPct = $DelayPct"
#if(!$PSBoundParameters.ContainsKey('ImagesCnt')) {
	#Write-Host "argument -ImagesCnt not used"
#}
#Write-Host "ImagesCnt = $ImagesCnt"
#Write-Host "Height = $Height"
#Write-Host "TrimBegPct = $TrimBegPct"
#Write-Host "TrimEndPct = $TrimEndPct"
#Write-Host ("TrimEndCust = " + ($TrimEndCust | ConvertTo-Json))
#Write-Host "FPS = $FPS"
#Write-Host "Quiet = $Quiet"
#Write-Host "Demo = $Demo"

#Set-StrictMode  -Version 3
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

# DEV NOTES: NOT CURRENTLY USED
# NOTE: Before you use the 'Recycle' module for the 1st time, 
#       you must execute the following command:
#       Set-PSRepository -Name 'PSGallery' -InstallationPolicy Trusted
# REFS:
#     1. https://stackoverflow.com/a/41985972/13499984
#     2. https://stackoverflow.com/questions/66385360/what-is-set-psrepository-scope-effect
# USAGE:
#     Remove-ItemSafely -LiteralPath "filespec"
# Install-Module -Name Recycle -Scope CurrentUser

# GATHER SCRIPT INFO 
$ScriptName = $MyInvocation.MyCommand.Name
$ScriptBaseName = $ScriptName.Replace('.ps1', '')
$ScriptPath = Split-Path $MyInvocation.MyCommand.Path

# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	return "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
}

# PARSE USER SETTINGS
$variable_test = Get-Variable -Name bg_msgs_on -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$bg_msgs_on = $true
}
$variable_test = Get-Variable -Name gif_images_cnt -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_images_cnt = 10
}
#Write-Host "gif_trim_beg = $gif_trim_beg"
$variable_test = Get-Variable -Name gif_trim_beg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_trim_beg = 0
}
#Write-Host "gif_trim_end = $gif_trim_end"
$variable_test = Get-Variable -Name gif_trim_end -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_trim_end = 0
}
$variable_test = Get-Variable -Name gif_trim_end_custom -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_trim_end_custom = @{}
}
#Write-Host ("gif_trim_end_custom = " + ($gif_trim_end_custom | ConvertTo-Json))
$variable_test = Get-Variable -Name gif_height -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_height = 320
}
$variable_test = Get-Variable -Name gif_fps -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_fps = 2
}
$variable_test = Get-Variable -Name gif_trim -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_trim = 0
}
$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$app_path = ""
} else {
	if (!$app_path.EndsWith('\')) {
		$app_path = $app_path + '\'
	}
}
$variable_test = Get-Variable -Name gif_archive_lifetime_hours -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_archive_lifetime_hours = 720
}
$variable_test = Get-Variable -Name bi_alert_folder_path -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$bi_alert_folder_path = "D:\BlueIris\Alerts"
}
$variable_test = Get-Variable -Name bi_new_folder_path -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$bi_new_folder_path = "D:\BlueIris\New"
}
$variable_test = Get-Variable -Name gif_DAT_reqd_images_cnt -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$gif_DAT_reqd_images = 4
}
$variable_test = Get-Variable -Name append_DAT_tag_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$append_DAT_tag_flg = $true
}
$variable_test = Get-Variable -Name dat_archive_lifetime_hours -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$dat_archive_lifetime_hours = 720
}
$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$post_to_BI_logfile_flg = $false
}
$variable_test = Get-Variable -Name post_all_events -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$post_all_events = $false
}

# INITIALIZE VARS
$PsVersion = $PSVersionTable.PSVersion.Major
# NOTE: Scripts running from Windows Task Scheduler have a parent process named 'svchost'.
#       https://stackoverflow.com/a/65894924
$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
# NOTE: Scripts called by another script have a 'PSCommandPath'.
#       Method 'PSCommandPath' returns null if script is executed via the terminal
#       https://stackoverflow.com/a/23370618/13499984
$CallingScript = $MyInvocation.PSCommandPath

# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM BLUE IRIS)
if ($ParentProcess -eq 'BlueIris') {
	$fromConsole = $false
} else {
	$fromConsole = ($CallingScript -eq $null)
}

# Get the current username and domain (hostname)
# NOTE: sometimes needed for msg.exe (e.g. from PowerShell ISE)
# ... tip: from the console, you can type 'hostname' or 'whoami'
$my_servername = (whoami).Split('\')[0]
$server_username = (whoami).Split('\')[1]

# For debugging
#msg.exe $server_username /TIME:10 "entered script '$ScriptName'"
#msg.exe $server_username /TIME:10 "PsVersion ($ScriptName) = $PsVersion"
#msg.exe $server_username /TIME:10 "ParentProcess ($ScriptName) = $ParentProcess"
#msg.exe $server_username /TIME:10 "fromConsole ($ScriptName) = $fromConsole"

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
if (!$fromConsole) {
	if ($bg_msgs_on) {
		$ExecPolicy = Get-ExecutionPolicy
		$exec_policy_str = "PSExecutionPolicy = $ExecPolicy"
		if ($env:PSExecutionPolicyPreference) {
			$ExecPolicyPref = $env:PSExecutionPolicyPreference
			$exec_policy_str = "Session ExecutionPolicy = $ExecPolicyPref"
		}
		$msg_sec = 10
		$bg_msg = "Script '$ScriptName' is running in the background" + 
		"`n$exec_policy_str" + 
		"`nPsVersion = $PsVersion" + 
		"`nParentProcess = $ParentProcess" + 
		"`nalert record = $AlertRec" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"
		msg.exe $server_username /TIME:$msg_sec $bg_msg
	}
}

# For testing CLI mode
#$fromConsole = $false

# ===============================================

function BIJsonCall() {
# PURPOSE: Get JSON interface response from Blue Iris web server.
#          Intended to be called by other PowerShell scripts.
# ARGS:    $JsonCmd         = (hashtable) like '@{cmd=xxx},@{cmd=xxx,camera=name},...'
#          $QuietFlg        = [OPTIONAL] (boolean) $true suppresses verbose output to the console
#          $FromConsoleFlg = [OPTIONAL] (boolean) $true notifies function that calling command was executed from the console
# RETURNS: (object) PowerShell PSCustomObject
#          or [OPTIONAL] (string) JSON (see Note #2)
# NOTES:   Use user setting 'return_json_flg' at your own risk. I've found the PSCustomObject to be MUCH more reliable.
# REFS:    1) Blue Iris Help PDF: chapter 'Administration', section 'JSON Interface'
#          2) https://www.houselogix.com/docs/blue-iris/BlueIris/json.htm  (old!)
#          3) https://ipcamtalk.com/threads/json-interface.21491/
#          4) https://ipcamtalk.com/threads/json-help-please.40381/#post-372770
#
# Version: X
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V1.0   - 2023-06-XX - initial release

	# MUST BE THE FIRST LINE IN A SCRIPT
	param (
		[hashtable]$JsonCmd,
		[boolean]$QuietFlg,
		[boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	$PsVersion = $PSVersionTable.PSVersion.Major

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "`Entered $fn_name()`n------------------------"
	#Write-Host ("$fn_name() JsonCmd = " + ($JsonCmd | ConvertTo-Json))
	#Write-Host "$fn_name() PsVersion = $PsVersion"	
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"	

	function encryptMD5($str) {
	# PURPOSE: Get MD5 hash for a string.
	# ARGS:    $str = (string) string to be encrypted
	# RETURNS: (string) MD5 encrypted string... like, '0753a0e20c9027ccab70f4291a21aa08'
	# REF:     https://infosecscout.com/get-md5-hash-in-powershell/
	# NOTES:   (from REF)
	#	Unfortunately, PowerShell provides no direct function to generate a hash from a string.
	#	However, PowerShell offers a cmdlet to generate MD5 hash for a file: Get-FileHash. 
	# 	The cmdlet can also be used to get the MD5 hash for a string, by opening a stream and hashing it.
	# UTILITY: https://www.md5hashgenerator.com/
	# ... Example: 'MD5string' returns '0753a0e20c9027ccab70f4291a21aa08'

		$stringAsStream = [System.IO.MemoryStream]::new()
		$writer = [System.IO.StreamWriter]::new($stringAsStream)
		$writer.write($str)
		$writer.Flush()
		$stringAsStream.Position = 0
		$hash_obj = Get-FileHash -InputStream $stringAsStream -Algorithm MD5
		
		# For debugging
		#$str = $hash_obj | Out-String
		#Write-Host "hash_obj = $str"

		return $hash_obj.Hash
	}

	# PARSE COMMANDLINE ARGS
	if ($JsonCmd -eq $null) {
		#if ($FromConsoleFlg -and !$QuietFlg) {
		if ($FromConsoleFlg) {
			Write-Host "ERR/expecting an argument: JsonCmd (PowerShell hashtable)`n"
			exit
		} else {
			return "ERR/expecting an argument: JsonCmd (PowerShell hashtable)"
		}
	} else {
		# For debugging
		#Write-Host ("requested JsonCmd type = " + ($JsonCmd.GetType()))
	}

	# PARSE OPTIONAL USER SETTINGS
	if (!(Test-Path variable:return_json_flg)) {
		$return_json_flg = $false  # Default
	}

	# ========================
	# PERFORM API CALLS

	# Make the URL
	$uri = "http://" + $script:bi_webserver + "/json";

	# For debugging
	#msg.exe $script:server_username /TIME:5 "uri = $uri"

	# SPECIAL HANDLING FOR 'geofence' JSON COMMAND REQUESTS
	$dev_uuid = $null  # Default
	if ($JsonCmd.uuid -ne $null) {
		$dev_uuid = $JsonCmd.uuid
	}

	# For debugging
	if ($dev_uuid) {
		#Write-Host "$fn_name() dev_uuid = $dev_uuid"
	}

	# EXECUTE API CALL
	
	# ================================================================
	# 1.1 PRELIMINARY LOGIN
	
	# Create, then encode the JSON 'login' command
	$parameters = @{
		cmd = 'login'
	}
	$json_params = $parameters | ConvertTo-Json

	# For debugging
	#msg.exe $script:server_username /TIME:15 "$fn_name() (1.1) login #1 json_params... `n$json_params"

	# Execute the JSON request
	$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

	# For debugging
	#msg.exe $script:server_username /TIME:15 "$fn_name() (1.1) login #1 jsonResponse... `n$jsonResponseObj"
	
	# Get the session id
	$session = $jsonResponseObj.session
	
	# ================================================================
	# 1.2 AUTHENTICATE THE LOGIN

	# Create MD5 hash-encoded credentials string
	$response_str = $bi_user_name + ":" + $session + ":" + $bi_user_pw
	$response_hash = encryptMD5 $response_str

	# Create, then encode a modified JSON 'login' command
	if ($dev_uuid) {
		$parameters = @{
			cmd = 'login'
			session = $session
			response = $response_hash
			uuid = $dev_uuid
		}
	} else {
		$parameters = @{
			cmd = 'login'
			session = $session
			response = $response_hash
		}
	}
	$json_params = $parameters | ConvertTo-Json

	# For debugging
	#Write-Host "$fn_name() (1.2) authorizing json login... `n$json_params"
	#msg.exe $script:server_username /TIME:15 "$fn_name() (1.2) authorizing json login ... `n$json_params"
	
	# Execute the JSON request
	$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

	# For debugging
	#Write-Host ("$fn_name() (1.2) jsonResponse... "`n + ($jsonResponseObj  | ConvertTo-Json))
	#msg.exe $script:server_username /TIME:15 "$fn_name() (1.2) jsonResponse... `n$jsonResponseObj"
	
	# Handle failure
	if ($jsonResponseObj.result -eq 'fail') {
		#if ($FromConsoleFlg -and !$QuietFlg) {
		if ($FromConsoleFlg) {
			Write-Host "ERR/unable to authenticate Blue Iris JSON interface`n"
			exit
		} else {
			return "ERR/unable to authenticate Blue Iris JSON interface"
		}
	}
	
	# Get the authenticated session id
	$auth_session_id = $jsonResponseObj.session

	# For debugging
	#msg.exe $script:server_username /TIME:5 "$fn_name() auth_session_id... $auth_session_id"

	# ================================================================
	# 1.3 EXECUTE THE REQUESTED JSON COMMAND

	if ($dev_uuid -eq $null) {
		# Modify, then encode the requested JSON login command
		$JsonCmd.session = $auth_session_id
		$json_params = $JsonCmd | ConvertTo-Json

		# For debugging
		#msg.exe $script:server_username /TIME:5 "$fn_name() (1.3) requested JsonCmd...`n$json_params"
		#Write-Host ("$fn_name() (1.3) requested JsonCmd...`n$json_params")

		$max_retries = 3 
		for ($i = 0; $i -le $max_retries; $i++) {

			# Execute the JSON request
			$params = @{
				Uri                = $uri
				#Headers            = @{"Content-Type"  = "application/json"}
				Method             = 'POST'
				Body               = $JsonCmd | ConvertTo-Json
				ContentType        = 'application/json'
			}
			
			# Alternate syntaxes
			$jsonResponseObj = Invoke-RestMethod @params
			#$jsonResponseObj = Invoke-RestMethod -Uri $uri -Method Post -Body ($JsonCmd | ConvertTo-Json) #>
			#$jsonResponseObj = ($JsonCmd | ConvertTo-Json) | Invoke-RestMethod -Uri $uri -Method Post

			if ($jsonResponseObj.result -eq 'fail') {
				#Write-Host ("$fn_name() JSON response for cmd '" + $JsonCmd.cmd + "' = `n" + ($jsonResponseObj | ConvertTo-Json))
			}

			if ($PsVersion -ge 7) {
				# Avoids the following WARNING...
				# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
				$jro = $jsonResponseObj | ConvertTo-Json -Depth 16
			} else {
				$jro = $jsonResponseObj | ConvertTo-Json
			}

			# For debugging
			#Write-Host "$fn_name() (1.3) jsonResponse... $jro"
			#msg.exe $script:server_username /TIME:15 "$fn_name() (1.3) JsonCmd Response... `n$jro"
			
			# Use loop to automatically retry for a limited set of times
			$retry_flg = (($jro -eq $null) -or $($jro -eq "null") -or $($jro -eq "") -or ($jro -eq "{}"))
			if ($retry_flg) {
				if ($i -eq $max_retries) {
					if ($FromConsoleFlg -and !$QuietFlg) {
						Write-Host "(quit retries after $max_retries iterations)"
					}
				}
				continue
			}
			break
		}
	}

	# For debugging
	if ($jsonResponseObj) {
		#Write-Host ("$fn_name() (1.3) jsonResponse type = " + $jsonResponseObj.GetType())
		#Write-Host ("$fn_name() (1.3) jsonResponse... `n" + ($jsonResponseObj  | ConvertTo-Json))
		#msg.exe $script:server_username /TIME:5 "$fn_name() (1.3) jsonResponse... $jsonResponseObj"
	} 

	# Retain for error testing
	#$jsonResponseObj = $null
	#$jsonResponseObj = ''

	if ($jsonResponseObj -eq $null) {
		return "ERR/null response from Blue Iris JSON interface"
	} 
	if ($jsonResponseObj -eq '') {
		return "ERR/empty response from Blue Iris JSON interface"
	}
	if ($jsonResponseObj.result -ne $null) {

		# For debugging
		#Write-Host ("$fn_name() (1.3) jsonResponseObj.result... " + ($jsonResponseObj.result))
		
		if ($jsonResponseObj.result -eq 'fail') {
			
			# Show the contents of the $jsonResponseObj 'data' key

			$json_data_obj = $jsonResponseObj.data
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) jsonResponseObj.data... " + ($jsonResponseObj.data))

			if ($PsVersion -ge 7) {
				# Avoids the following WARNING...
				# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
				$jdo_str = $json_data_obj | ConvertTo-Json -Depth 16
			} else {
				$jdo_str = $json_data_obj | ConvertTo-Json
			}
			
			# 'Minify' the json string.
			if ($jdo_str) {
				if (!($jdo_str -match 'null')) {
					#$jdo_str = ( (ConvertFrom-Json $jdo_str) | ConvertTo-Json -Compress )
					#Write-Host "jdo = $jdo_str"
					return "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is 'result:fail'"
				} else {					
					#return ("ERR/JSON response.data for JSON cmd '" + $JsonCmd.cmd + "' is null/empty")
					return ("ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is null")
				}
			}
		}
	}
	if ($return_json_flg) {
		if ($PsVersion -ge 7) {
			# Avoids the following WARNING...
			# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
			return ($jsonResponseObj | ConvertTo-Json -Depth 32)
		} else {
			return ($jsonResponseObj | ConvertTo-Json)
		}
	} else {
		return $jsonResponseObj
	}
}

function BIJsonClipstats() {
# UTILITY TO GET BLUE IRIS JSON 'CLIPSTATS' CMD PROPERTIES
#
# Description of JSON cmd 'clipstats' from the Blue Iris Help PDF:
# "return information about a specific database record."
#
# Version: X
# Created by jaydeel at ipcamtalk.com
#
# USAGE:      BIJsonClipstats -Clips list -Props list -Quiet
# ARGUMENTS:  $Clips           = [REQUIRED] (string) comma-delimited QUOTED clipnames, like "@926894649","@926871636"
#                                 NOTE: extensions accepted, but not required, e.g., "@926894649.bvr",
#             Props            = [OPTIONAL] (string) comma-delimited JSON properties, like 'name,push,count'
#             $QuietFlg        = [OPTIONAL] (boolean) true suppresses verbose informational output to the console
#             $FromConsoleFlg = [OPTIONAL] (boolean) $true notifies function that calling command was executed from the console
# RETURNS:    1. (ordered hashtable) if multiple clips and/or properties requested
#             2. (string) if a single clip and a single property requested
#             3  'NA' if the clip is specified and the property does not exists
# USAGE EXAMPLE
#		$response_obj = BIJsonClipstats -Clips $AlertRec -Props 'camera,path,file,msec,alertmsec,res'
#
#       From the response, set these variables:
#		$clipstats_result = $response_obj[0] 
#                         like, {"result": "success"}, or {"result": "fail"}
#		$clipstats_data = $response_obj[1]
#                         like {"@152866668.bvr": "NA"}
#		                  or   {"@152866668.bvr": {
#			                       "camera": "DW1",
#			                       "path": "@152720932.bvr",
#			                       "file": "DW1.20230530_060000.1965063.17-0.jpg",
#			                       "msec": 3599166,
#			                       "alertmsec": 27426,
#			                       "res": "2688x1520"
#			                       "dummy": "NA"
#		                        }

	# MUST BE THE FIRST LINE IN A SCRIPT
	param(
		[string]$Clips,
		[string]$Props,
		[boolean]$QuietFlg,
		[boolean]$FromConsoleFlg = $false
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() Clips = '$Clips'"
	#Write-Host "$fn_name() Props = '$Props'"
	
	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$Clips) {
		$err_msg = "ERR/expecting an argument (-Clips)"
		return $err_msg
	}
	$my_clips_ary = $Clips.split(',')
	
	# HANDLE OPTIONAL: COMMANDLINE ARGS
	$my_props_ary = @()
	if ($Props) {
		$my_props_ary = $Props.split(',')
	}

	# For debugging
	#Write-Host ("$fn_name() my_clips_ary = " + $my_clips_ary)
	#Write-Host ("$fn_name() my_clips_ary cnt = " + $my_clips_ary.length)
	#Write-Host ("$fn_name() my_props_ary = " + $my_props_ary)

	# INITIALIZE
	$rtn_obj = [ordered]@{}
	<# $rtn_obj = [ordered]@{
		success = $null
		data = $null
	} #>
	$x = ""

	if ($my_clips_ary.length -gt 0) {
		$i = 1;
		foreach ($clip in $my_clips_ary) {
			
			if ($FromConsoleFlg -and !$QuietFlg) {
				if ($i -eq 1) {
					#Write-Host "$fn_name() processing clip '$clip'"
				} else {
					#Write-Host "`n$fn_name() processing clip '$clip'"
				}
			}

			# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
			# ... REQUIRES COMPANION SCRIPT 'BI_json_call.ps1'
			#     WHICH RETURNS A POWERSHELL PSCustomObject
			$jsonCmd = @{
				cmd = 'clipstats'
				path = $clip
			}
			
			#$json_response_obj = BIJsonCall -JsonCmd $jsonCmd -QuietFlg $QuietFlg
			$json_response_obj = `
				BIJsonCall `
				-JsonCmd $jsonCmd `
				-QuietFlg $QuietFlg `
				-FromConsoleFlg $FromConsoleFlg

			# For debugging
			#Write-Host ("$fn_name() json_response_obj type = " + $json_response_obj.GetType())
			#Write-Host ("$fn_name() json_response_obj** = " + ($json_response_obj | ConvertTo-Json))
			#Write-Host ("json_response_obj.result** = " + ($json_response_obj.result))

			# If GET failed, process next clip
			if ($json_response_obj -match 'ERR/') {
				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host $json_response_obj
				}
				$rtn_obj[$clip] = 'NA'
				break
			}
			if ($json_response_obj.result -eq "fail") {

				# for debugging
				#Write-Host ("$fn_name() json_response_obj.result type = " + (($json_response_obj.result).GetType()))
				#Write-Host ("j$fn_name() son_response_obj.result = " + ($json_response_obj.result | ConvertTo-Json))

				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host "$fn_name() json_response_obj.result = fail"
					Write-Host ($json_response_obj | ConvertTo-Json)
				}
				$rtn_obj[$clip] = 'NA'
				break
			}

			# Extract data from the JSON command response
			$data_obj = $json_response_obj.data

			# For debugging
			#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
			#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))

			# $data_obj is a PSCustomObject like
			# @{
			#	camera=DW1;
			#	path=@746553655.bvr;
			#	file=DW1.20230428_220000.1572042.17-0.jpg;
			#	offset=1572009;
			#	date=1682744400;
			#	color=16711808;
			#	flags=134217985;
			#	zones=0;
			#	memo=;
			#	plate=;
			#	res=2688x1520;
			#	msec=2814017;
			#	alertmsec=24520:
			#	camprofile=7;
			#	camschedule=1;
			#	camsmart=2;
			#	camzones=1;
			#	campreset=30;
			#	filesize=46m55s (419.5M);
			#	filetype=bvr H264 New
			# }

			# Convert PSCustomObject into a hashtable
			# REF: https://tinyurl.com/ddatr635
			$hashtable = @{}
			foreach ($property in $data_obj.psobject.properties.name) {
				$hashtable[$property] = $data_obj.$property
			}
			$data_obj_keys = $hashtable.Keys

			# For debugging
			#Write-Host "$fn_name() hashtable = $hashtable"
			#Write-Host "$fn_name() data_obj_keys = $data_obj_keys"

			$new_obj = [ordered]@{}		
			if ($my_props_ary.length -gt 0) {
				# LIST REQUESTED PROPERTIES
				foreach ($clipProp in $my_props_ary) {
					if ($data_obj.$clipProp -or $data_obj.$clipProp -eq 0) {
						if ($FromConsoleFlg -and !$QuietFlg) {
							#Write-Host ($clipProp + " = "  + $data_obj.$clipProp)
						}
						$new_obj.Add($clipProp, $data_obj.$clipProp)
						$x = $data_obj.$clipProp
					} else {
						if ($FromConsoleFlg -and !$QuietFlg) {
							#Write-Host ($clipProp + " = NA")
						}							
						$new_obj.Add($clipProp, "NA")
						$x = "NA"
					}
				}
				$rtn_obj[$clip] = $new_obj
				
			} else {
				# LIST ALL PROPERTIES
				foreach ($clip_prop in $data_obj_keys) {
					if ($data_obj.$clip_prop -or $data_obj.$clip_prop -eq 0) {
						if ($FromConsoleFlg -and !$QuietFlg) {
							#Write-Host ($clip_prop + " = "  + $data_obj.$clip_prop)
						}							
						$new_obj.Add($clip_prop, $data_obj.$clip_prop)
					} else {
						if ($FromConsoleFlg -and !$QuietFlg) {
							#Write-Host ($clip_prop + " = NA")
						}							
						$new_obj.Add($clip_prop, "NA")
					}
				}
				$rtn_obj[$clip] = $new_obj
			}
			$i++
		}
	} else {
		$data_obj_keys = @()
		foreach ($item in $data_obj) {
			$data_obj_keys += $($item | Get-Member -MemberType *Property).Name
		}
		foreach ($key in $data_obj_keys) {
			if ($FromConsoleFlg -and !$QuietFlg) {
				#Write-Host ($key + " = "  + $data_obj.$key)					
			}
			$rtn_obj.Add($key, $data_obj.$key)
		}
	}

	# Handle single clip, single property request
	if ($my_clips_ary.length -eq 1 -and $my_props_ary.length -eq 1) {
		$rtn_obj = $x
	}

	# LOG OUT OF API SESSION
	$jsonCmd = @{cmd = 'logout'}
	#BIJsonCall -JsonCmd $jsonCmd -QuietFlg $QuietFlg
	BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg
								
	if ($FromConsoleFlg -and !$QuietFlg) {
		Write-Host "Clipstats:"
		Write-Host "----------"
		Write-Host ($rtn_obj | ConvertTo-Json)
	}

	return $rtn_obj
}

function FindBytes() {
# Soruce: https://stackoverflow.com/a/62511302/13499984
	param (
		[byte[]]$Bytes, 
		[byte[]]$Search, 
		[int]$Start, 
		[Switch]$All		
	)
    for ($idx = $Start; $idx -le $Bytes.Length - $Search.Length ; $idx++) {
        for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
        if ($i -ge $Search.Length) { 
            $idx
            if (!$All) { 
				return
			}
        } 
    }
}

function RepairDatFile() {
# PURPOSE: Try to repair a DAT file with unexpected structure [unbalanced JPG beg/end markers]
# ARGS:    see param()
# NOTE:    Repair is attempted only when the count of 'end markers' <> the count of 'beg markers'
# RETURNS  [array] @([boolean]rtn_flg, [array]SOI_mrks, [array]EOI_mrks)

	param (
		[byte[]]$Bytes,
		[array]$BegOfImageMarks,
		[array]$EndOfImageMarks
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	#Write-Host "Entered $fn_name()..."
	#Write-Host ("$fn_name() BegOfImageMarks type = " + $BegOfImageMarks.GetType())
	#Write-Host ("$fn_name() BegOfImageMarks cnt = " + $BegOfImageMarks.Count)
	#Write-Host "$fn_name() BegOfImageMarks = $BegOfImageMarks"
	#Write-Host ("$fn_name() EndOfImageMarks type = " + $EndOfImageMarks.GetType())
	#Write-Host ("$fn_name() EndOfImageMarks cnt = " + $EndOfImageMarks.Count)
	#Write-Host "$fn_name() EndOfImageMarks = $EndOfImageMarks"
	
	# Initialize vars
	$rtn_flg = $true  # Assume success
	$SOI_mrks_cnt = $BegOfImageMarks.Count
	$EOI_mrks_cnt = $EndOfImageMarks.Count		

	# Must convert arrays to ArrayLists to use method 'RemoveAt()' (later)
	$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$BegOfImageMarks
	$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EndOfImageMarks

	# For debugging
	#Write-Host ("$fn_name() SOI_mrks_lst type = " + $SOI_mrks_lst.GetType())
	#Write-Host ("$fn_name() SOI_mrks_lst cnt = " + $SOI_mrks_lst.Count)
	#Write-Host "$fn_name() SOI_mrks_lst = $SOI_mrks_lst"
	#Write-Host ("$fn_name() EOI_mrks_lst type = " + $EOI_mrks_lst.GetType())
	#Write-Host ("$fn_name() EOI_mrks_lst cnt = " + $EOI_mrks_lst.Count)
	
	if ($SOI_mrks_cnt -gt $EOI_mrks_cnt) {

		for ($i=0; $i -lt $EOI_mrks_cnt; $i++) {
			$indices_delta = [int]$EndOfImageMarks[$i] - [int]$BegOfImageMarks[$i]
			
			# For debugging
			#Write-Host ("- end index = [$i] " + $EndOfImageMarks[$i])
			#Write-Host "- indices_delta = $indices_delta"

			if ($indices_delta -lt 0) {
				
				$SOI_mrks_lst_new = $SOI_mrks_lst
				$SOI_mrks_lst_new.RemoveAt($i)

				# For debugging
				#Write-Host "starting recursive call..."
				#Write-Host "- SOI_mrks_lst_new = $SOI_mrks_lst_new"

				# Recursive call
				$arr1 = [array]$SOI_mrks_lst_new
				$arr2 = [array]$EOI_mrks_lst
				$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
					RepairDatFile `
					-Bytes $bytes `
					-BegOfImageMarks $arr1 `
					-EndOfImageMarks $arr2

				# For debugging
				#Write-Host "recursive result_flg = $result_flg"

				if ($result_flg) {
					$rtn_flg = $true
					$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
					$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
					break
				} else {
					$rtn_flg = $false
				}				
			}
		}					
	}
	
	if ($EOI_mrks_cnt -ge $SOI_mrks_cnt) {

		for ($i=0; $i -lt $SOI_mrks_cnt; $i++) {
			$indices_delta = [int]$EndOfImageMarks[$i] - [int]$BegOfImageMarks[$i]
			
			# For debugging
			#Write-Host ("- beg index = [$i] " + $BegOfImageMarks[$i])
			#Write-Host "- indices_delta = $indices_delta"
			
			if ($indices_delta -lt 0) {
				
				$EOI_mrks_lst_new = $EOI_mrks_lst
				$EOI_mrks_lst_new.RemoveAt($i)

				# For debugging
				#Write-Host "starting recursive call..."
				#Write-Host "- EOI_mrks_lst_new = $EOI_mrks_lst_new"

				# Recursive call
				$arr1 = [array]$SOI_mrks_lst
				$arr2 = [array]$EOI_mrks_lst_new
				$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
					RepairDatFile `
					-Bytes $bytes `
					-BegOfImageMarks $arr1 `
					-EndOfImageMarks $arr2
					
				# For debugging
				#Write-Host "recursive result_flg = $result_flg"

				if ($result_flg) {
					$rtn_flg = $true
					$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
					$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
					break
				} else {
					$rtn_flg = $false
				}				
			}
		}
	}

	return @($rtn_flg, [array]$SOI_mrks_lst, [array]$EOI_mrks_lst)

}

function IsJpegFileValid {
# PURPOSE: To confirm that a JPG file is a valid JPG image
# NOTES:
#    Using 2 methods:
#    1. Test image type
#       REFS:  http://learningpcs.blogspot.com/2011/07/powershell-validate-jpeg-files.html
#              https://www.powershellgallery.com/packages/PScribo/0.9.1/Content/Src%5CPrivate%5CGet-ImageMimeType.ps1
#    2. Per the JPEG Interchange Format... any JPEG file or stream that is written according to the 
#       JPEG Interchange Format means that they must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) 
#       and end with the EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
#       These two markers will not be found anywhere else in a JPEG file/stream.
#       REFS: https://stackoverflow.com/a/46809226
#             https://www.codeproject.com/Questions/421855/Verifying-a-file-is-JPEG
# DISCUSSION:
#    Originally I used only Method #1 in this function. However, I learned later that Method #1 would still return valid
#    if the image bytes did not terminate with the EOI (End-of-Image) marker (0xFF, 0xD9). These JPG files still opened 
#    just fine in every windows viewer I tried (Photos, Paint, Paint3D, PhotoScape X, ACDSee...), HOWEVER, if they were 
#    opened in a web browser (tested Chrome, Edge) the images had a white band across the bottom. 
#    Additionally, web apps like https://app.roboflow.com/ would not import the JPGs.
#    So clearly, Method 1 is not sufficient. For redundancy, I've opted to use both methods. 

	param(
		[string]$FilePath
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging.
	#msg.exe $script:server_username /TIME:15 "entered function $fn_name()..."
	#Write-Host "entered function $fn_name()..."
	#Write-Host "$fn_name() FilePath = '$FilePath'"
	#msg.exe $script:server_username /TIME:30 "$fn_name() FilePath = '$FilePath'"
	
	# Initialize flags
	$method1_flg = $false
	$method2_flg = $false
	
	# ===========================
	# METHOD 1
	# Validate Image file Type

	# WARNING: The following code line IS ESSENTIAL when executing from Task Scheduler (or Blue Iris)
	#          (however works just fine from Powershell ISE)
	#          REF: https://tinyurl.com/2p929nhz
	add-type -AssemblyName System.Drawing
	
	try {
		$img = [System.Drawing.Image]::FromFile("$FilePath");
		$method1_flg = $img.RawFormat.Equals([System.Drawing.Imaging.ImageFormat]::Jpeg)
		$img.Dispose()  # Release resources

	} catch [OutOfMemoryException] {
		$method1_flg = $false
	}

	# ===========================
	# METHOD 2
	# Validate bytes per the 'JPEG Interchange Format' (see Notes above)
	
	# Import the file to a byte[] array
	$ImageBytes = [System.IO.File]::ReadAllBytes($FilePath)
	
	# Capture the start and end byte pairs.
	#$bytes_rng1 = $ImageBytes[0],$ImageBytes[1]
	#$bytes_rng2 = $ImageBytes[$ImageBytes.Count-2],$ImageBytes[$ImageBytes.Count-1]
	$bytes_rng1 = $ImageBytes[0..1]
	$bytes_rng2 = $ImageBytes[($ImageBytes.Count-2)..($ImageBytes.Count-1)]

	# Define the required JPG start/end markers
	$markers1 = @(255,216) # hexArray @("FF", "D8")
	$markers2 = @(255,217) # hexArray @("FF", "D9")
	<# $hexArray = "FF", "D8"
	$markers1 = @()
	foreach ($hex in $hexArray) {
		$markers1 += [Int32]"0x$hex"
	}
	$hexArray = "FF", "D9"
	$markers2 = @()
	foreach ($hex in $hexArray) {
		$markers2 += [Int32]"0x$hex"
	} #>

	# Validate the start and end byte pairs.
	# ... Note: Method Compare-Object returns null if the arrays are the identical
	$x1 = Compare-Object $bytes_rng1 $markers1
	$x2 = Compare-Object $bytes_rng2 $markers2
	if ($x1 -eq $null -and $x2 -eq $null) {
		$method2_flg = $true
	} else {
		$method2_flg = $false
	}
	
	# For testing
	#$method2_flg = $true

	# For debugging
	#Write-Host "markers1 = $markers1, markers2 = $markers2"
	#Write-Host "bytes_rng1 = $bytes_rng1, bytes_rng2 = $bytes_rng2"
	#Write-Host "method1_flg = $method1_flg, method2_flg = $method2_flg"
	#msg.exe $script:server_username /TIME:15 "$fn_name() method1_flg = $method1_flg, method2_flg = $method2_flg"
	
	if ($method1_flg -and $method2_flg) {
		return $true
	} else {
		return $false
	}
}

function ExtractDatImages() {
# PURPOSE:  Extract all JPG files from a Blue Iris *.DAT file, and save them to disk
#           with filenames like 'img01.jpg', 'img02.jpg', ...
# ARGS:     see param lines below
# RETURNS:  1. (hashtable) @{ source=filespec ; data=@(jpgfilelist) }
#           2. or (string) if warning/error (NOTE: string must begin with 'WARNING:' or 'ERR/')
# NOTES:    1. TBD
#
# ATTRIBUTIONS:
#    Per Ken @blueirissupport... "The DAT file is custom serialized format. 
#    To extract JPGs: search the byte content for blocks that start with 'FF D8 FF E0' and go through 'FF D9'. 
#    To extract the JSON string: after the bitmap there are 2 QWORDs (8 bytes each), then the STRING for the JSON 
#    (comprised of a DWORD length then the characters)."
#
#    Per the JPEG Interchange Format... any JPEG file or stream is written according to the JPEG Interchange 
#    Format must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) and end with the 
#    EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
#    These two markers will not be found anywhere else in a JPEG file/stream.
#    REF: https://stackoverflow.com/a/46809226
#
# KEY POINT: 
#    JPG images in a BI *.DAT file are stored in reverse chronological order
	
    param(
		[string]$AlertFile,          # [REQUIRED] Alert file, like 'GT.20230610_070000.41437.17-0.jpg'
		[string]$AlertPath,          # [REQUIRED] folder path for Blue Iris 'DAT' files (full path)
		[string]$DestPath,           # [REQUIRED] folder for the extracted JPG files
		[int]$DatReqdImages,         # number of images that must be found before extracting any images	
		[int]$ImagesCnt,             # number of images to be extracted from the BVR file, if the DAT file determined to be 'bad'	
		[boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
		[boolean]$NukeDatFileFlg,    # true removes DAT file from Blue Iris storage (typ. if used solely to extract JPGs)
		[boolean]$NoArchiveDatFlg,   # true suppresses the archiving of Blue Iris DAT file (if used to create GIF)
		[boolean]$QuietFlg,          # true suppresses verbose output to the console
		[boolean]$FromConsoleFlg     # true notifies function that calling command was executed from the console
)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	$PsVersion = $PSVersionTable.PSVersion.Major

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() AlertFile = '$AlertFile'"
	#Write-Host "$fn_name() AlertPath = '$AlertPath'"
	#Write-Host "$fn_name() DestPath = '$DestPath'"
	#Write-Host ("$fn_name() DatReqdImages type = " + $DatReqdImages.GetType())
	#Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
	#Write-Host "$fn_name() NukeDatFileFlg = $NukeDatFileFlg"
	#Write-Host "$fn_name() NoArchiveDatFlg = $NoArchiveDatFlg"
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"
	#Write-Host "$fn_name() PsVersion = $PsVersion"	
	
	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$AlertFile) {
		$err_msg = "ERR/expecting an argument (-AlertFile -DestPath)"
		return $err_msg
	}
	if (!$DestPath) {
		$err_msg = "ERR/expecting an argument (-AlertFile -DestPath)"
		return $err_msg
	} else {
		$DestPath = $DestPath.TrimEnd('\')
	}
	
	# Initialize
	$bkup_path = "$DestPath\_dat_archive"
	$rtn_obj = [ordered]@{}
	$arr = $AlertFile.Split(".")
	$cam = $arr[0]
	$marker = $arr[1]
	
	# For debugging
	#Write-Host "$fn_name() arr = '$arr'"
	
	# TEST FOR DESTINATION FOLDER FOR JPGs
	$jpg_path =  $DestPath + "\" + $cam
	if (!(Test-Path -LiteralPath "$jpg_path")) {
		New-Item -ItemType Directory -Force -Path $jpg_path
	}
	
	# For debugging
	#Write-Host "$fn_name() jpg_path = '$jpg_path'"

	# DELETE EXISTING *.JPG FILES ASYNCHRONOUSLY
	get-childitem -path ($jpg_path + "\img*.jpg") | remove-item -force
	
	# =====================================================================
	# TRY TO GET A 'DAT' FILE ... 2 METHODS
	# 1. LOOK FOR OLD 'DAT' FILE
	# 2. GET MOST RECENT 'DAT' FILE
	
	# METHOD #1 - LOOK FOR OLD 'DAT' FILE
	#    Try to use $AlertFile as a pattern (DAT files have the same basename as $AlertFile)
	$dat_file = $AlertPath + "\" + $AlertFile.Replace(".jpg", ".dat")
	$dat_file_found_flg = Test-Path -LiteralPath "$dat_file"

	# For debugging
	#msg.exe $script:server_username /TIME:60 "$fn_name() `ndat_file1_found_flg = $dat_file_found_flg`ndat_file = '$dat_file'"
	
	# METHOD #1 - GET MOST RECENT 'DAT' FILE
	#    Try to get the camera's most recently created 'DAT' file
	if (!$dat_file_found_flg) {
		
		# For debugging
		#msg.exe $script:server_username /TIME:60 "$fn_name() trying to get most recent DAT file"

		$pattern = $AlertPath + "\" + $cam + ".*.dat"
		if ($PsVersion -ge 7) {
			$dat_file = (Get-ChildItem $pattern | Sort -Descending) | Select -First 1
		} else {
			$dat_file = ((Get-ChildItem $pattern).FullName | Sort -Descending) | sSelect -First 1
		}
		
		$max_age_sec = 60
		if ($dat_file) {
			$file_dt = (Get-Item -LiteralPath "$dat_file").LastWriteTime
			$diff = (Get-Date).Subtract($file_dt)
			$age_sec = [math]::Floor($diff.TotalSeconds)

			# For debugging
			#Write-Host "DAT file is $age_sec old ('$dat_file')"
			
			if ($age_sec -lt $max_age_sec) {
				$dat_file_found_flg = $true
			}
		}

		# For debugging
		#msg.exe $script:server_username /TIME:60 "$fn_name() dat_file2_found_flg = $dat_file_found_flg`ndat_file = '$dat_file'`nage_sec = $age_sec"

	}

	# PROCEED IF 'DAT' FILE FOUND
	if ($dat_file_found_flg) {
		
		if ($FromConsoleFlg -and !$QuietFlg) {
			Write-Host "`nAlert AI 'DAT' File Processing Summary:"
			Write-Host "---------------------------------------"
			Write-Host "Found DAT file: '$dat_file'"
		}
	
		# LOAD THE 'DAT' FILE
		if ($PsVersion -ge 7) {
			#$bytes = [System.IO.File]::ReadAllBytes("$dat_file")
			# Ref: https://stackoverflow.com/a/63782608/13499984
			$bytes = Get-Content -LiteralPath "$dat_file" -AsByteStream -Raw
		} else {
			$bytes = Get-Content -LiteralPath "$dat_file" -Encoding Byte -Raw
		}

		# For debugging
		#Write-Host ("$fn_name() bytes len = " + $bytes.Count)
		#Write-Host $bytes
		#$msg = $bytes.Count
		#msg.exe $script:server_username /TIME:60 "$fn_name() bytes len = $msg"
		
		# JPG BYTE MARKER NEEDLES
		# ... JPGs start
		$hexArray = "FF", "D8", "FF", "E0"
		$search_ary1 = @()
		foreach ($hex in $hexArray) {
			$search_ary1 += [Int32]"0x$hex"
		}
		# ... JPGs end
		$hexArray = "FF", "D9"
		$search_ary2 = @()
		foreach ($hex in $hexArray) {
			$search_ary2 += [Int32]"0x$hex"
		}

		# For debugging.
		#Write-Host "$fn_name() jpg data starting byte pattern = $search_ary1"
		#Write-Host "$fn_name() jpg data ending byte byte pattern = $search_ary2"
		
		$found_ary1 = FindBytes -Bytes $bytes -Search $search_ary1 -All
		$found_ary2 = FindBytes -Bytes $bytes -Search $search_ary2 -All
		$found_ary1_cnt = $found_ary1.Count
		$found_ary2_cnt = $found_ary2.Count

		# For debugging.
		#Write-Host "$fn_name() jpg data start indices = $found_ary1"
		#Write-Host "$fn_name() jpg data end indices = $found_ary2"
		#msg.exe $script:server_username /TIME:60 "$fn_name()... `nfound_ary1_cnt = $found_ary1_cnt `nfound_ary2_cnt = $found_ary2_cnt"
		
		# Validate expected DAT images count
		$bad_DAT_file_flg = $false
		if ($found_ary1_cnt -lt $DatReqdImages) {
			$bad_DAT_file_flg = $true
			$bad_DAT_msg = "DAT file has fewer images than required ($found_ary1_cnt/$DatReqdImages)... "
			$bad_DAT_msg += "`n   Please navigate to Blue Iris 'Camera settings' > 'Trigger' tab > 'Artificial Intelligence' dialog"
			$bad_DAT_msg += "`n   - Confirm that setting 'To cancel' has a value like DoNotCancel"
			$bad_DAT_msg += "`n   - Confirm that setting '+ real-time images' has a value = $DatReqdImages (or " + ($DatReqdImages-1) + " if using motion-leading image)"
			$bad_DAT_msg += "`nTrying to extract images from BVR instead..."
		}
		# Validate expected DAT structure
		# ... expect $found_ary1 like '25 113202'
		# ...        $found_ary2 like '113054 227880'
		if (!$bad_DAT_file_flg) {
			$i = 0
			foreach ($jpg_idx in $found_ary1) {
				if ($i -gt 0) {
					if ([int]$jpg_idx -ge [int]$found_ary2[$i - 1]) {
						$bad_DAT_file_flg = $true
						$bad_DAT_msg = "DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
						break
					}
				}
			}
		}
		
		# For debugging
		#Write-Host "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"
		#msg.exe $script:server_username /TIME:60 "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"

		$extract_success_flg = $false  # Assume failure

		if (!$bad_DAT_file_flg) {

			# DAT file has at least one JPG file
			if ($found_ary1_cnt -gt 0) {

				# DAT file structure is NOT as expected
				if ($found_ary1_cnt -ne $found_ary2_cnt) {
				
					# Try to repair DAT file marker arrays before giving up
					$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
						RepairDatFile `
						-Bytes $bytes `
						-BegOfImageMarks $found_ary1 `
						-EndOfImageMarks $found_ary2
						
					# For debugging
					#Write-Host "result_flg = $result_flg"

					if ($result_flg) {
						$found_ary1 = $SOI_mrk_idxs_ary
						$found_ary2 = $EOI_mrk_idxs_ary
						$found_ary1_cnt = $SOI_mrk_idxs_ary.Count
						$found_ary2_cnt = $EOI_mrk_idxs_ary.Count

						if ($FromConsoleFlg) {
							if ($found_ary1_cnt -gt $found_ary2_cnt) {
								Write-Host "REPAIRED 'beg marker' indices: now = $img_SOI_mrk_idxs"
							} else {
								Write-Host "REPAIRED 'end marker' indices; now = $img_EOI_mrk_idxs"
							}
						}
					}
				}

				# DAT file structure is as expected
				if ($found_ary1_cnt -eq $found_ary2_cnt) {

					if ($found_ary1_cnt -ge $DatReqdImages) {
			
						if ($ImagesCntUsedFlg) {
							if ($FromConsoleFlg) {
								Write-Warning "Argument '-ImagesCnt' ignored; it applies only to GIFs created from BVR files"
							}
						}

						$rtn_obj['source'] = $dat_file
						$jpgs = @()
						$jpgs_cnt = 0
						$i = 0
						foreach ($item in $found_ary1) {
							
							$bounding_idx1 = $item
							$bounding_idx2 = $found_ary2[$i] + 1  # Include the EOI marker!
							$img_bytes = $bytes[$bounding_idx1..$bounding_idx2]
							
							if ($img_bytes.Count -gt 0) {	

								# Build JPG's filespec
								$jpg_filename_idx = $found_ary1_cnt - $i
								$jpg_filename_sfx = ([string]$jpg_filename_idx).PadLeft(2,'0')
								$jpg_filespec = "$jpg_path\img$jpg_filename_sfx.jpg"
							
								# Save the JPG file to disk					
								if ($PsVersion -ge 6) {
									# REF: https://stackoverflow.com/a/63782608/13499984
									Set-Content "$jpg_filespec" -Value $img_bytes -AsByteStream 
								} else {
									Set-Content "$jpg_filespec" -Value $img_bytes -Encoding Byte
								}

								# Confirm that the extracted data is a valid JPG file
								$jpg_valid_flg = `
									IsJpegFileValid `
									-FilePath $jpg_filespec

								# For debugging
								#msg.exe $script:server_username /TIME:60 "$jpg_filespec is valid = $jpg_valid_flg"
								
								if (!$jpg_valid_flg) {
									if ($FromConsoleFlg) {
										Write-Warning "DAT file export aborted because '$jpg_filespec' is NOT a valid JPG"
									}
									break
								} else {
									$jpgs_cnt++
									$jpgs += $jpg_filespec
								
									# For debugging
									#Write-Host "$fn_name() jpgs = $jpgs"
									#Write-Host "'$jpg_filespec' is a valid JPG"
								}
							}
							$i++
						}

						$rtn_obj['data'] = $jpgs
						
						if ($jpgs.Count -gt 0) {
							$extract_success_flg = $true
						}

						if ($FromConsoleFlg -and !$QuietFlg) {
							Write-Host ("Saved $jpgs_cnt JPG files to folder '$jpg_path'")
						}
						
					} else {
						$rtn_obj = "NO ACTION: DAT file contains only $found_ary1_cnt/$DatReqdImages required JPGs"							
					}			

				} else {
					$rtn_obj = "WARNING: DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
				}
			} else {
				$rtn_obj = "WARNING: No JPGs found in DAT file '$dat_file'"			
			}		
		} else {
			$rtn_obj = "WARNING: " + $bad_DAT_msg
		}
	} else {
		if ($FromConsoleFlg -and !$QuietFlg) {
			Write-Host "Unable to find DAT file '$dat_file'"
		}
	}
	
	# For debugging
	#msg.exe $script:server_username /TIME:60 "extract_success_flg = $extract_success_flg"
	

	# DELETE THE DAT FILE (if requested)
	if ($NukeDatFileFlg) {
		if ($dat_file_found_flg -and $extract_success_flg) {
			
			#       *** WARNING ***
			# When this script is executed from a background process, the attempted Recycle Bin 
			# command PERMANENTLY deletes the specified file (see 'VisualBasic:DeleteFile' action). 
			# 
			# To get around this, we ALWAYS back up to the files to an alternate folder for safe keeping.
			# To prevent this folder from growing excessively large, use user-setting $dat_archive_lifetime_hours.
			
			# HANDLE EXCEPTION FOR BACKGROUND PROCESSES
			# ...  'Task Scheduler' or 'Blue Iris'		
			#$process = $script:ParentProcess
			#$background_processes = @('svchost', 'BlueIris')
			#$background_process_flg = ($background_processes -contains $process)
			$background_process_flg = !$FromConsoleFlg
						
			# For debugging
			if (!$FromConsoleFlg) {
				#msg.exe $script:server_username /TIME:60 "nuking DAT file from background process: '$process'"
			}

			if (!$NoArchiveDatFlg) {
				# BACK UP THE DAT FILE TO APP-MANAGED STORAGE
				$bkup_path = "$DestPath\_dat_archive"
				if (!(Test-Path -LiteralPath "$bkup_path")) {
					$null = New-Item -ItemType Directory -Force -Path $bkup_path
				}
				Copy-Item -LiteralPath "$dat_file" -destination $("$bkup_path") -force
			}

			# MOVE THE FILE FROM BLUE IRIS STORAGE TO THE WINDOWS RECYCLE BIN
			# OR PERMANENTLY DELETE THE FILE IF RUNNING FROM BACKGROUND
			Add-Type -AssemblyName Microsoft.VisualBasic
			[Microsoft.VisualBasic.FileIO.FileSystem]::DeleteFile("$dat_file", 'OnlyErrorDialogs', 'SendToRecycleBin')

			if ($FromConsoleFlg -and $NoArchiveDAtFlg) {				
				# PERMANENTLY DELETE THE FILE FROM BLUE IRIS STORAGE
				Remove-Item -LiteralPath "$dat_file"
				#Write-Host "NOTE: DAT (permanently) deleted at user's request"
			}

			if ($FromConsoleFlg) {
				if (!$QuietFlg) {
					Write-Host "NOTE: DAT removed from Blue Iris storage at user's request (to Recycle Bin & '$bkup_path')"
				}
			} else {
				$bg_msg = "NOTE: DAT removed from Blue Iris storage at user's request (to '$bkup_path')"
				#msg.exe $script:server_username /TIME:15 $bg_msg
			}
		}
	}
		
	return $rtn_obj
}

function MakeGIF() {
# PURPOSE:  Make GIF image file to be used as attachment for Pushover notification 
# ARGS:     see param lines below
# RETURNS:  1. (hashtable) @{file=$output_fs delay=$delay_ms ; cnt=$ImagesCnt ; height=$Height ; fps=$FPS ; kb=$gif_kb ; 
#                            msec=$gif_dur_ms ; aldur_ms=$al_dur_ms ; res=$gif_jpg_wxh ; 
#                            cam=$al_cam ; alfile=$al_file ; archive=
#                          }
#           2. or (string) if error
# NOTES:    1. Be sure to check if 'gif_size' = 0, as this indicates the file was not actually created
#           2. Pushover API attachment maximum size is 2,621,440 bytes (2621 KB)
# ATTRIBUTIONS:
#    Thanks to bp2008 for providing clues to Blue Iris HTTP command to extract JPGs from a BVR
	
    param(
		[string]$RootPath,           # [REQUIRED] root folder used by this function (full path)
		[string]$BvrPath,            # [REQUIRED] folder path for Blue Iris' 'BVR' files (full path)
		[string]$AlertPath,          # [REQUIRED] folder path for Blue Iris' 'DAT' files (full path)
		[string]$AlertRec,           # Alert record, like '@123456789'
		[string]$AlertMemo,          # alert AI memo, typ. via Blue Iris macro &MEMO; like 'vehicle:90%,person:65%'
		[boolean]$LiveFlg,           # create GIF from specified camera's stream
		[string]$Camera,             # camera shortname (required with -Live argument)
		[string]$DbRec,              # hyphenated numeric values following UI3 URL argument '&rec=', like 'ui3.htm?rec=208953322-1181690'
		[int]$Seconds,               # GIF duration (-1 uses alert's duration via JSON cmd 'clipstats'
		[boolean]$UseDatFile,        # extract JPGs from AI DAT if it exists
		[int]$DatReqdImages,         # number of images that must be found before extracting any images	
		[boolean]$NoPushFlg,         # true prevents limiting GIF size to below 2.5MB
		[boolean]$NukeDatFileFlg,    # true removes DAT file from Blue Iris storage (typ. if used solely to extract JPGs)
		[boolean]$NoArchiveDatFlg,   # true suppresses the archiving of Blue Iris DAT file (if used to create GIF)
		[boolean]$NoArchiveGifFlg,   # true suppresses the archiving of created GIF file
		[int]$DelayPct,              # percentage of alert's duration to wait before extracting JPGs from clip
		[int]$ImagesCnt,             # number of JPG frames to extract from BVR
		[boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
		[int]$Height,                # height of image extracted from BVR in pixels
		[int]$FPS,                   # GIF file frames per second
		[int]$TrimBegPct,            # trim beginning of alert clip by this percentage
		[int]$TrimEndPct,            # trim end of alert clip by this percentage
		[hashtable]$TrimEndCust,     # adjust $TrimEndPct by custom amount for AI labels in user-setting $gif_trim_end_custom
		[boolean]$QuietFlg,          # true suppresses verbose output to the console
		[boolean]$FromConsoleFlg     # $true notifies function that calling command was executed from the console
)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"	
	#Write-Host "$fn_name() RootPath = '$RootPath'"
	#Write-Host "$fn_name() BvrPath = '$BvrPath'"
	#Write-Host "$fn_name() AlertPath = '$AlertPath'"
	#Write-Host "$fn_name() AlertRec = '$AlertRec'"
	#Write-Host "$fn_name() AlertMemo = '$AlertMemo'"
	#Write-Host "$fn_name() LiveFlg = '$LiveFlg'"
	#Write-Host "$fn_name() Camera = '$Camera'"
	#Write-Host "$fn_name() UI3Rec = '$DbRec'"
	#Write-Host "$fn_name() Seconds = $Seconds"
	#Write-Host ("$fn_name() Seconds = " + $Seconds.GetType())
	#Write-Host "$fn_name() RootPath = '$RootPath'"
	#Write-Host "$fn_name() UseDatFile = $UseDatFile"
	#Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
	#Write-Host "$fn_name() NukeDatFileFlg = $NukeDatFileFlg"
	#Write-Host "$fn_name() NoArchiveDatFlg = $NoArchiveDatFlg"
	#Write-Host "$fn_name() DelayPct = $DelayPct"
	#Write-Host "$fn_name() ImagesCnt = $ImagesCnt"
	#Write-Host ("$fn_name() ImagesCntUsedFlg type = " + $ImagesCntUsedFlg.GetType())
	#Write-Host "$fn_name() ImagesCntUsedFlg = $ImagesCntUsedFlg"
	#Write-Host "$fn_name() Height = $Height"
	#Write-Host "$fn_name() FPS = $FPS"
	#Write-Host ("$fn_name() TrimBegPct type = " + $TrimBegPct.GetType())
	#Write-Host "$fn_name() TrimBegPct = $TrimBegPct"
	#Write-Host "$fn_name() TrimEndPct = $TrimEndPct"
	#Write-Host ("$fn_name() TrimEndCust = " + ($TrimEndCust | ConvertTo-Json))
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"

	# For debugging
	$bg_msg =  "Entered function: $fn_name()"
	#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
	
	# HANDLE COMMANDLINE ARGS
	if ($LiveFlg) {

		# Override default or requested settings
		$UseDatFile = $false
		
		# Get the MOST RECENT BVR FILE for the specified camera
		$bvr_timestamp = Get-Date -Format "yyyyMMdd_"
		#$bvr_offset = Get-Date -Format "yyyyMMdd_"
		$pattern = "$BvrPath\$Camera.$bvr_timestamp*bvr"
		$bvr_file = (Get-ChildItem $pattern | Sort -Descending) | Select -First 1
		<# if ($PsVersion -ge 7) {
			$bvr_file = (Get-ChildItem $pattern | Sort -Descending) | Select -First 1
		} else {
			$bvr_file = ((Get-ChildItem $pattern).FullName | Sort -Descending) | Select -First 1
		} #>
		
		if ($bvr_file) {
			
			# CALCULATE THE OFFSET TIME - NOW - based the most recent BVR file's creation time
			$file_dt = (Get-Item -LiteralPath "$bvr_file").CreationTime
			$diff = (Get-Date).Subtract($file_dt)
			$bvr_offset_ms = [math]::Floor($diff.TotalMilliSeconds)

			# Build the filename for archived files
			$bvr_filename = $bvr_file.Name
			$arr = ($bvr_file.Name).Split('.')
			$bvr_path = $arr[0] + "." + $arr[1] + ".bvr"
			$archive_bvr_path = $arr[0] + "." + $arr[1] + "." + $bvr_offset_ms + ".bvr"
			
			# For debugging
			#Write-Host ("bvr_file arr = " + ($arr | ConvertTo-Json))
			#Write-Host "bvr_path = '$bvr_path'"
			#Write-Host "bvr_offset_ms = '$bvr_offset_ms'"
			#Write-Host "archive_bvr_path = '$archive_bvr_path'"
			
			if ($FromConsoleFlg -and !$Quiet) {
				Write-Host "Active bvr_file: '$bvr_file'"
				Write-Host "Active bvr_file offset: $bvr_offset_ms msec"
			}
			# For debugging
			$bg_msg = "$fn_name() Active bvr_file:`n'$bvr_file'`nbvr_file offset: $bvr_offset_ms ms"
			#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on

		} else {
			$err_msg = "ERR/unable to get active BVR filename: '$bvr_file'"
			return $err_msg
		}
	}
	
	if ($DbRec) {
		
		# Also supports UI3 style 'rec'
		# Example: http://192.168.1.3:81/ui3.htm?rec=213168393-3165057&cam=GT
		# UI3Rec = '213168393-3165057'
		
		$hypen_idx = $DbRec.IndexOf('-')
		if ($hypen_idx) {
			$arr = $DbRec.Split('-')
			$AlertRec = "@" + ($arr[0]).ToString()
			
			# Currently not used
			if ($arr[1]) {
				$dbrec_offset = [int]$arr[1]
			} else {
				$dbrec_offset = $null
			}
			
			# For debugging
			#Write-Host "Database record ID = $AlertRec"
			#Write-Host "Database record offset = $dbrec_offset"
			
		} else {
			$AlertRec = "@" + $DbRec.ToString()
			$dbrec_offset = $null
		}
	}
	
	if ($DelayPct -lt 0) {
		$err_msg = "ERR/argument '-DelayPct' must be greater than zero"
		return $err_msg
	}
	if ($ImagesCnt -le 0) {
		$err_msg = "ERR/argument '-ImagesCnt' must be greater than zero"
		return $err_msg
	}
	if ($Height -le 0) {
		$err_msg = "ERR/argument '-Height' must be greater than zero"
		return $err_msg
	}
	if ($FPS -le 0) {
		$err_msg = "ERR/argument '-FPS' must be greater than zero"
		return $err_msg
	}
	if (($TrimBegPct -lt 0) -or ($TrimBegPct -ge 99)) {
		$err_msg = "ERR/argument '-TrimBegPct' must be between 0 and 99"
		return $err_msg
	}
	if (($TrimEndPct -lt 0) -or ($TrimEndPct -ge 99)) {
		$err_msg = "ERR/argument '-TrimEndPct' must be between 0 and 99"
		return $err_msg
	}
	
	$gif_trim_end_cust_keys = $gif_trim_end_custom.Keys
	#Write-Host "gif_trim_end_cust_keys = $gif_trim_end_cust_keys"
	foreach ($cust_lbl in $gif_trim_end_cust_keys) {
		$cust_lbl_pct = $gif_trim_end_custom[$cust_lbl]
		$cust_lbl_disp = "gif_trim_end_custom[$cust_lbl]"
		$clip_dur_adj = $TrimBegPct + $TrimEndPct + $cust_lbl_pct
		#Write-Host "clip_dur_adj ($cust_lbl) = $clip_dur_adj"
		if ($clip_dur_adj -gt 100) {
			$err_msg = "ERR/sum of clip duration adjustments cannot exceed <100% ('TrimBegPct/TrimEndPct/$cust_lbl_disp = $TrimBegPct/$TrimEndPct/$cust_lbl_pct)"
			return $err_msg
		} 
		if ($clip_dur_adj -lt 0) {
			$err_msg = "ERR/sum of clip duration adjustments cannot be <0% ('TrimBegPct/TrimEndPct/$cust_lbl_disp = $TrimBegPct/$TrimEndPct/$cust_lbl_pct)"
			return $err_msg
		} 
	}
	
	# Initialize
	$gif_method = "viaClip"  # Default
	$delay_ms = 0
	$gif_images_cnt = $ImagesCnt
	$gif_dur_adj_pct = 0	
	
	# TRY TO CREATE GIF
	# ... up to 3x before giving up
	for ($i=1; $i -le 3; $i++) {
		
		if ($i -gt 0) {
			#Write-Host "$fn_name() iteration #$i"
		}

		# Adjust GIF images count for alerts with memos that start with 'vehicle'-like terms
		# ... Typically memos are like 'vehicle:78%,person:52%'
		# ... After BI version 5.5.7, memos may be like 'RuleEngine/LineDetector/Crossed; vehicle:78%,person:52%'
		$arr = $AlertMemo.Split(';')
		$al_memo_ai_str = ""
		foreach ($item in $arr) {
			if ($item.IndexOf('%') -gt 0) {
				$al_memo_ai_str = $item.Trim()
				break
			}
		}
		if ($al_memo_ai_str.length -gt 0) {
			$al_memo_ai_str_pcs = $al_memo_ai_str.Split(',')
			$j = 0
			foreach ($al_memo_ai_str_pc in $al_memo_ai_str_pcs) {
				$arr = $al_memo_ai_str_pc.Split(':')
				$al_memo_ai_str_pcs[$j] = $arr[0]
				$j++
			}
			$first_lbl = $al_memo_ai_str_pcs[0]

			if ($TrimEndCust[$first_lbl]) {
				$gif_dur_adj_pct = $TrimEndCust[$first_lbl]
			}

			# For debugging
			#Write-Host "$fn_name() al_memo_ai_str_pcs = $al_memo_ai_str_pcs"
			#Write-Host "$fn_name() first_lbl = $first_lbl"
			#Write-Host "$fn_name() gif_dur_adj_pct = $gif_dur_adj_pct"
		}

		if (!$LiveFlg) {

			# GET THE ALERT'S CLIP DETAILS
			$clip_stats_obj = `
				BIJsonClipstats `
				-Clips $AlertRec `
				-Props 'camera,path,file,offset,msec,alertmsec,res' `
				-FromConsoleFlg $FromConsoleFlg `
				-Quiet $QuietFlg 

			# Expecting response_obj to be a hashtable, like ... 
			#[
			#	{
			#		"result": "success"
			#	},
			#	{
			#		"@152866668.bvr": {
			#			"camera": "DW1",
			#			"path": "@152720932.bvr",
			#			"file": "DW1.20230530_060000.1965063.17-0.jpg",
			#			"offset": 2016632,
			#			"msec": 3599166,
			#			"alertmsec": 27426,
			#			"res": "2688x1520"
			#		}
			#	}
			#]

			if ($clip_stats_obj.GetType() -match "string") {
				if ($clip_stats_obj.StartsWith("ERR/")) {
					$err_msg = $clip_stats_obj
					return $err_msg
				}
			}
			
			# For debugging
			#Write-Host ("$fn_name() response_obj type = " + $clip_stats_obj.GetType())
			#Write-Host ("$fn_name() response_obj = " + ($clip_stats_obj | ConvertTo-Json))

			$clipstats_result = $clip_stats_obj[0]  # A single element array
			$clipstats_data = $clip_stats_obj[1]  # An array of arrays
			$clipstats_success_flg = ($clipstats_result -match "result=success")
			
			# For debugging
			#Write-Host "$fn_name() clipstats_result = $clipstats_result"
			#Write-Host ("$fn_name() clipstats_data = " + ($clipstats_data | ConvertTo-Json))
		}
		
		if ($clipstats_success_flg -or $LiveFlg) {

			if (!$LiveFlg) {
				
				# Because BIJsonClipstats() accepts multiple AlertRec's,
				# we need to use only the 1st item
				$item = $clipstats_data[0]
			
				# For debugging
				#Write-Host ("$fn_name() item = " + ($item | ConvertTo-Json))

				if ($item -match 'NA') {
				
					# GET ALERT'S DETAILS
					# ... $item via BIJsonClipstats -Clips $AlertRec -Props 'camera,path,file,msec,alertmsec,res'
					$al_cam     = $item.camera
					$al_path    = $item.path
					$al_file    = $item.file
					$al_res     = $item.res  # WxH
					$al_offset  = [int]$item.offset
					if ($Seconds -gt 0) {
						$al_dur_ms = $Seconds * 1000
					} else {
						$al_dur_ms = [int]$item.alertmsec
					}
					if ($al_file) {
						$arr = $al_file.Split('.')
						$bvr_file = $BvrPath + "\" + $arr[0] + "." + $arr[1] + ".bvr"
					} else {
						$err_msg = "ERR/non-existent database record id"
						return $err_msg									
					}
					
					# For debugging
					#Write-Host "bvr_file = $bvr_file"
					
					# SPECIAL CASE...
					# ... Handle occasional instances when 'clipstats' yields a zero value for 'alertmsec'
					if ($DbRec) {
						if ($al_dur_ms -eq 0) {
							$err_msg = "ERR/argument '-AlertRec', 'AlertDB', '-DbRec', or '-UI3Rec' does not point to an EXISTING Alert record"
							return $err_msg			
						}
					}

					# Adjust duration
					$gif_dur_ms = [math]::Floor( $al_dur_ms - ($TrimBegPct/100)*$al_dur_ms - ($TrimEndPct/100)*$al_dur_ms - ($gif_dur_adj_pct/100)*$al_dur_ms )

					# Adjust source clip offset
					$clip_offset = $al_offset
					if ($TrimBegPct -gt 0) {
						$gif_offset = $clip_offset + ($TrimBegPct/100)*$al_dur_ms
					} else {
						$gif_offset = $clip_offset
					}
					
					if ($gif_dur_ms -lt $al_dur_ms) {
						if ($FromConsoleFlg -and !$QuietFlg) {
							Write-Host "`nGIF Duration Mods:"
							Write-Host "------------------"
							Write-Host "TrimBegPct/TrimEndPct/OtherAdjPct = $TrimBegPCt/$TrimEndPCt/$gif_dur_adj_pct"
							Write-Host "Alert's duration = $al_dur_ms msec"
							Write-Host "GIF's duration   = $gif_dur_ms msec"
						}
					}
		
					# Calculate time delta between each extacted BVR image
					$step_ms = [math]::floor($gif_dur_ms / $gif_images_cnt)

					# For debugging
					#Write-Host "$fn_name() al_cam = $al_cam"
					#Write-Host "$fn_name() al_path = $al_path"
					#Write-Host "$fn_name() al_offset = $al_offset"
					#Write-Host "$fn_name() gif_offset = $gif_offset"
					#Write-Host "$fn_name() al_dur_ms = $al_dur_ms"
					#Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"
					#Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
					#Write-Host "$fn_name() step_ms = $step_ms"
					
				} else {
					$err_msg = "ERR/non-existent alert record '$AlertRec'"
					return $err_msg
				}
			} else {

				$al_cam    = $Camera
				$al_path   = $bvr_path
				$al_file   = $bvr_filename
				$al_offset = $bvr_offset_ms
				$al_dur_ms = $Seconds * 1000
				$al_res    = "WxH"
				$delay_ms  = 0

				# For debugging
				#Write-Host "$fn_name() al_cam = $al_cam"

				# For debugging
				#rite-Host "al_cam = $al_cam"
				
				# Adjust source clip duration
				$gif_dur_ms = [math]::Floor( $al_dur_ms - ($TrimBegPct/100)*$al_dur_ms - ($TrimEndPct/100)*$al_dur_ms - ($gif_dur_adj_pct/100)*$al_dur_ms )

				# Adjust source clip offset
				$clip_offset = $al_offset
				if ($TrimBegPct -gt 0) {
					$gif_offset = $clip_offset + ($TrimBegPct/100)*$al_dur_ms
				} else {
					$gif_offset = $clip_offset
				}
	
				# Calculate time delta between each extracted BVR image
				$step_ms = [math]::floor($gif_dur_ms / $gif_images_cnt)
				#$step_ms = $gif_dur_ms / $gif_images_cnt

			}				
			
			# MAKE SUBFOLDERS FOR JPGS, GIFS, DATS
			# ... Note: $RootPath is user-defined (_user_settings.ps1 file)
			$path = $RootPath + $al_cam  # like 'D:\pushover_gifs\DW1' 
			if (!(Test-Path -LiteralPath $path)) {
				New-Item -ItemType Directory -Force -Path $path
			}
			$gif_archive_path = ($RootPath + "_gif_archive")  # like 'D:\pushover_gifs\_gif_archive'
			if (!(Test-Path -LiteralPath "$gif_archive_path")) {
				New-Item -ItemType Directory -Force -Path $gif_archive_path
			}
			$dat_archive_path = ($RootPath + "_dat_archive")  # like 'D:\pushover_gifs\_dat_archive'
			if (!(Test-Path -LiteralPath "$dat_archive_path")) {
				New-Item -ItemType Directory -Force -Path $dat_archive_path
			}
		
			# DELETE EXISTING *.JPG AND *.GIF FILES ASYNCHRONOUSLY
			get-childitem -path ($path + "\img*.jpg") | remove-item -force
			get-childitem -path ($path + "\*.gif") | remove-item -force

			# TRY GETTING GIF JPGs 2 WAYS
			# ==============================================================
			# 1. GET JPGs FROM ALERT'S DAT FILE (IF IT EXISTS)
			$dat_jpgs_cnt = 0
			if ($UseDatFile) {
				if ($FromConsoleFlg -and !$QuietFlg) {
					#Write-Host ""
					#Write-Host "Attempting to extract JPGs from AI Analysis 'DAT' file..."
				}
				
				# Extract JPGs
				$extract_obj = `
					ExtractDatImages `
					-AlertFile $al_file `
					-AlertPath $bi_alert_folder_path `
					-DestPath $RootPath `
					-DatReqdImages $DatReqdImages `
					-ImagesCnt $ImagesCnt `
					-ImagesCntUsedFlg $ImagesCntUsedFlg `
					-NukeDatFileFlg $NukeDatFileFlg `
					-NoArchiveDatFlg $NoArchiveDatFlg `
					-Quiet $QuietFlg `
					-FromConsoleFlg $FromConsoleFlg
				
				# For debugging
				#Write-Host ("$fn_name() ExtractDatImages extract_obj = " + ($extract_obj | ConvertTo-Json))
				#$msg = "extract_obj type = " + $extract_obj.GetType()
				#msg.exe $script:server_username /TIME:60 $msg
				#msg.exe $script:server_username /TIME:60 $extract_obj.Count

				if ($extract_obj) {
					if ($extract_obj -match 'ERR/' -or $extract_obj -match 'WARNING:') {
						if ($FromConsoleFlg) {
							$extract_obj = $extract_obj.Replace("WARNING: ", "")
							Write-Warning $extract_obj
						}
					} elseif ($extract_obj.Count -gt 0) {
						$gif_method = "viaDat"
						$dat_file = $extract_obj['source']
						$dat_jpgs_cnt = ($extract_obj['data']).Count
						$gif_images_cnt = $dat_jpgs_cnt
						if ($FromConsoleFlg -or $QuietFlg) {
							#Write-Host $extract_obj
						}
					} else {
						# Do nothing
					}
				}
			}
			
			# For debugging
			#msg.exe $script:server_username /TIME:60 "gif_method = $gif_method"

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# 2. GET JPGs FROM ALERT'S VIDEO CLIP FILE
			if (!($gif_method -match "viaDat")) {

				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host "`nGetting JPGs from Blue Iris Clip..."
					Write-Host "-----------------------------------"
				}

				# For debugging
				$bg_msg = "$fn_name()`nGetting JPGs from Blue Iris Clip..."
				#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
				
				# BUILD ARRAYS FOR A RUNSPACE THAT WILL EXECUTE MANY URLS 
				$urls = @()   # Blue Iris HTTP Interface commands - to extract single JPGs from an alert's BVR
				$files = @()  # JPG filenames, serialized like 'img001.jpg' to 'img010.jpg' 
				$offset = $gif_offset
				for ($j=0; $j -lt $gif_images_cnt; $j++) {

					# Build the URL
					$time = $offset + ($j * $step_ms)
					$uri = $script:bi_webserver + "/file/clips/" + $al_path + "?time=" + $time + "&cache=1&h=" + $Height

					# Build the full filespec for the URL's output
					$jpg_filename_sfx = ([string]($j + 1)).PadLeft(2,'0')
					$filespec = $RootPath + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"

					$urls  += $uri
					$files += $filespec
				}

				# For debugging
				#Write-Host ("urls cnt = " + $urls.Count)
				#Write-Host ("files = `n" + ($files | Out-String))
				
				# KEY POINT; A Delay is necessary so that HTTP '/clips' commands successfully retrieve current images
				# My testing suggests that this delay must be at least as long as the duration of the alert clip we are using.
				# REF: In response to an email inquiry regarding this observation, Blue Iris support (Ken) replied
				#     "If recording continuous or with the "continuous +" modes the recording is delayed by the duration of the pre trigger buffer."
				if (!$DbRec) {
					$delay_ms = $gif_dur_ms*($DelayPct/100)
					#Write-Host "delay_ms = $delay_ms"
				}
				
				# For testing
				#$delay_ms = 5500
				
				if ($delay_ms -gt 0) {
					$gif_method = "viaClipDelayed"				
					#if ($FromConsoleFlg -and !$QuietFlg) {
						#Write-Host "waiting for clip buffering: $delay_ms msec..."
					#}
					if ($fromConsole) {
						$delay_sec = $delay_ms / 1000
						$activity_str = "Retrieving JPGs..."
						$sleep_msg = "waiting for clip buffering ($delay_sec seconds)..."
						SleepWithStatusBar -Seconds ($delay_ms  / 1000) -Activity $activity_str -Message $sleep_msg
					} else {
						Start-Sleep -Milliseconds $delay_ms
					}
				}

				# CALL LIST OF URLS ASYNCHRONOUSLY
				# REF: https://stackoverflow.com/a/31525044/13499984

				$maxConcurrentJobs = $gif_images_cnt
				$content = $urls

				# Create a RUNSPACE POOL where $maxConcurrentJobs is the 
				# maximum number of runspaces allowed to run concurrently    
				$RunspacePool = [runspacefactory]::CreateRunspacePool(1,$maxConcurrentJobs)

				# Open the runspace pool (very important)
				$RunspacePool.Open()

				# Provide feedback for a noticeably slow process
				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host "Extracting JPGs from video clip file..."
				}

				# Initialize collectors
				$ps_instances = @()  
				$jobs = New-Object System.Collections.ArrayList
				
				$j = 0
				foreach ($url in $content) {
					
					$file = $files[$j]

					# For debugging
					$bg_msg = "$fn_name()`nfile = $file"
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
					
					# Create a new PowerShell instance
					# REF: https://devblogs.microsoft.com/scripting/beginning-use-of-powershell-runspaces
					$ps_instance = [powershell]::Create()
					
					# Add the instance to a collector (to facilitate disposal later)
					$ps_instances += $ps_instance

					# Tell the instance to execute in our runspace pool
					$ps_instance.RunspacePool = $RunspacePool

					# Attach some code to the PS instance
					# ... "Invoke-RestMethod is perfect for quick APIs that have no special response information such as Headers or Status Codes, 
					#     whereas Invoke-WebRequest gives you full access to the Response object and all the details it provides."
					#$ps_instance.AddCommand("Invoke-WebRequest").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file)
					$ps_instance.AddCommand("Invoke-RestMethod").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file)
					
					# Begin execution asynchronously
					#$ps_instance.BeginInvoke()
					#$ps_instance.BeginInvoke() | Out-Null
					$handle = $ps_instance.BeginInvoke()

					# Fill collector
					$job_obj = [PSCustomObject]@{
						instance = $null
						data = $null
					}
					$job_obj.instance = $ps_instance
					$job_obj.data = $handle
					$jobs.Add($job_obj)
	
					# Give feedback on progress
					if ($FromConsoleFlg -and !$QuietFlg) {
						#Write-Host ("Initiated request for {0}" -f $url)
						Write-Host "Initiated request for $url"
						#Write-Host ("Initiated request for $url`n" + ($handle | ConvertTo-Json))
						#Write-Host ("handle = " + ($handle | ConvertTo-Json -Depth 16))
						#Write-Host ("Available Runspaces in RunspacePool: {0}" -f $RunspacePool.GetAvailableRunspaces())
						<# Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
							$_.data.iscompleted -ne 'Completed'
						}).Count) #>
					}
					
					# For debugging
					#$bg_msg = ("$fn_name()`nInitiated request for $url`n" + ($handle | ConvertTo-Json))
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
					
					$j++
				}

				#Write-Host ("Jobs: " + ($jobs | ConvertTo-Json))

				# WAIT UNTIL ALL JPG FILES CREATED
				$jpg_filename_sfx = ([string]$gif_images_cnt).PadLeft(2,'0')
				$filespec = $RootPath + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"
				$iter = 0
				$max_tries = 20
				$success_flg = $true
				$sleep_ms = 25
				do { 

					# Check status of jobs in runspace pool
					<# Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
						$_.data.iscompleted -ne 'Completed'
					}).Count) #>

					if (-not (Test-Path $filespec)) {
						Start-Sleep -Milliseconds $sleep_ms
						$iter++
					}
					if ($iter -gt $max_tries) {
						$success_flg = $false
						break
					}
				} until (Test-Path $filespec)
			
				# Feedback async status
				if ($success_flg) {
					$ascync_msg = "SUCCESS: $gif_images_cnt asynchronous web requests; Final JPG ready after $iter $sleep_ms-msec waits ($filespec)"
				} else {
					$ascync_msg = "FAIL: $gif_images_cnt asynchronous web requests; Final JPG NOT ready after $iter $sleep_ms-msec waits ($filespec)"
				}			
				if ($FromConsoleFlg -and !$QuietFlg) {
					Write-Host $ascync_msg
				}
				# For debugging
				#PopupBackgroundMsg -Seconds 30 -Message $ascync_msg -UserName $server_username -InBackground $bg_msgs_on
				
				#Write-Host (“Available Runspaces in RunspacePool: {0}” -f $RunspacePool.GetAvailableRunspaces())

				# Perform a proper cleanup to avoid possible resources issues / memory leaks
				<# $return = $job_obj | ForEach {
					$_.instance.EndInvoke($_.handle)
					$_.instance.Dispose()
				}
				$job_obj.clear() #>
				foreach ($ps_instance in $ps_instances) {
					$ps_instance.Dispose()  # Release resources
				}
				$ps_instances.Clear()
			}
					
			# ==============================================================
			# CREATE THE GIF FILE

			# But first, try to get the JPG file dimensions
			# ... Why limit attempts to 3x? 
			#     because BVR extraction via HTTP '/files/clip/' is occasionally unsuccessful
			$gif_jpg_WxH = "NA"  # Default
			for ($tries = 1; $tries -le 3; $tries++) {
				$jpg_file = $RootPath + $al_cam + "\img0" + $tries + ".jpg"
				if (Test-Path -LiteralPath "$jpg_file") {
					$img = [System.Drawing.Image]::FromFile($jpg_file); 
					$gif_jpg_WxH = "$($img.Width)x$($img.Height)"
					$img.Dispose()  # Release resources
					break
				}
			}
			
			# For debugging
			#Write-Host "gif_jpg_WxH = $gif_jpg_WxH"

			$input_fs  = $RootPath + $al_cam + "\img%02d.jpg"
			$output_fs = $RootPath + $al_cam + "\output.gif"
			$jpg_file_idx = $gif_images_cnt
			$i = 0
			do { 
			
				# Try removing one JPG at a time to ensure GIF is below maximum size 
				# accepted by the Pushover API (2.5MB)  ... https://pushover.net/api#attachments
				if ($i -gt 0) {
					$jpg_filename_sfx = ([string]$i).PadLeft(2,'0')	
					$jpg_remove_fs = $RootPath + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"
					if (Test-Path -LiteralPath "$jpg_remove_fs") {
						Remove-Item -LiteralPath "$jpg_remove_fs"
						$gif_images_cnt--
					} else {
						# Give up
						break
					}
				}
				
				if ($FromConsoleFlg -and !$Quietflg) {
					Write-Host "`nExecuting ffmpeg:"
					Write-Host "------------------"
				}
			
				<# ffmpeg `
					-hide_banner `
					-loglevel error `
					-nostats `
					-y `
					-f image2 `
					-framerate $FPS `
					-i $input_fs `
					$output_fs #>

				ffmpeg `
					-hide_banner `
					-loglevel error `
					-y `
					-f image2 `
					-framerate $FPS `
					-i $input_fs `
					$output_fs
				
				# For debugging
				#$bg_msg = "$fn_name()`noutput_fs = $output_fs"
				#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on

				# FFMPEG RESPONSES I'VE SEEN
				# [mjpeg @ xxxxxxxx] overread 8
				#   1. "This is an error that is thrown when ffmpeg tries to read a JPEG image (or a frame from an MJPEG stream) 
				#      block-by-block, but there are still leftover bits."
				#      REF:  https://superuser.com/a/559996
				#   2. "could not open file: I/O Error"
				#      REF:  TBD
				#   3. "Error demuxing input file 0: I/O Error..."
				#      REF:  TBD
				#   4. blah, blah.
				#      REF:  TBD			
				
				if (!$output_fs) {
					$err_msg = "ERR/ffmpeg could not create the GIF file"
					return $err_msg
				}

				# Get size of GIF file
				# ... DEPRECATED: waiting for the file to be created appears to be unnecessary
				<# $gif_kb = 0
				$iter = 0
				$max_tries = 20
				$success_flg = $true
				do { 
					if (-not (Test-Path $output_fs)) {
						Start-Sleep -Milliseconds 50
						$iter++
					}
					if ($iter -gt $max_tries) {
						$success_flg = $false
						break
					}
				} until (Test-Path $output_fs) #>
				
				if (Test-Path -LiteralPath "$output_fs") {
					$sizeKB = (Get-Item -Path $output_fs).Length/1kB
				} else {
					$sizeKB = 'NA'
				}
				
				if ($NoPushFlg) {
					# No need to limit GIF size if we're not using the Pushover API
					break
				}
				
				# Reduce the image count, but not below 1
				if ($i -gt 0) {
					$ImagesCnt--
					if ($ImagesCnt -lt 2) {
						break
					}
				}
				
				$i++
				
			} until ($sizeKB -lt 2500)
			
			$gif_kb = [double]([math]::Round($sizeKB,1).ToString("#.0"))
			
			# ARCHIVE THE GIF FILE
			if ($LiveFlg) {
				$gif_archive_fn = $archive_bvr_path.Replace(".bvr",".gif")			
			} else {
				$gif_archive_fn = $al_file.Replace(".jpg",".gif")
			}
			$gif_archive_fs = $gif_archive_path  + "\" + $gif_archive_fn			
			if (!(Test-Path -LiteralPath "$gif_archive_fs")) {
				#Copy-Item -path $output_fs -destination $gif_archive_fs -Force
				Copy-Item -path $output_fs -destination $gif_archive_fs
				$archive_gif_fs = $gif_archive_fs
			} else {
				# Save multiple copies, like 'filename(N).gif'
				$archive_gif_fs = `
					CopyFileAsSequence `
					-SourceFilespec $output_fs `
					-TargetFilespec $gif_archive_fs `
					-DestPath $gif_archive_path
			}
			
			# For debugging
			#Write-Host "gif_archive_fs = $gif_archive_fs"

			
			if ($FromConsoleFlg -and !$QuietFlg) {
				Write-Host "`nArchived Files Maintenance Summary:"
				Write-Host "-----------------------------------"
			}

			# PERFORM MAINTENANCE ON GIF ARCHIVE FOLDER
			if ($gif_archive_lifetime_hours -gt 0) {
				$gif_purge_filespec = $gif_archive_path  + "\*.gif"
				$arr = `
					PurgeOldFiles `
					-Pattern $gif_purge_filespec `
					-Hours $gif_archive_lifetime_hours
					
				# For debugging
				#Write-Host "gif_purge_filespec = '$gif_purge_filespec'"
				#Write-Host ("GIF purge arr = " + ($arr | ConvertTo-Json))
				
				if ($FromConsoleFlg -and !$QuietFlg) {
					if ($arr[1] -gt 0) {			
						Write-Host ("count of archived GIF files matching '$gif_purge_filespec' = " + $arr[0])
						Write-Host ("count of archived GIF files purged (age >= $gif_archive_lifetime_hours hrs) = " + $arr[1])
					} else {
						Write-Host "No old GIF files found to purge (age >= $gif_archive_lifetime_hours hrs)"
					}
				}
			}

			# PERFORM MAINTENANCE ON DAT ARCHIVE FOLDER
			if ($dat_archive_lifetime_hours -gt 0) {
				$dat_purge_filespec = $dat_archive_path  + "\*.dat"
				$arr = `
					PurgeOldFiles `
					-Pattern $dat_purge_filespec `
					-Hours $dat_archive_lifetime_hours
					
				# For debugging
				#Write-Host "dat_purge_filespec = '$dat_purge_filespec'"
				#Write-Host ("DAT purge arr = " + ($arr | ConvertTo-Json))
				
				if ($FromConsoleFlg -and !$QuietFlg) {
					if ($arr[1] -gt 0) {			
						Write-Host ("count of archived DAT files matching '$dat_purge_filespec' = " + $arr[0])
						Write-Host ("count of archived dat_archive_path files purged (age >= $dat_archive_lifetime_hours hrs) = " + $arr[1])
					} else {
						Write-Host "No old DAT files found to purge (age >= $dat_archive_lifetime_hours hrs)"
					}
				}
			}
			
		} else {
			# TRY AGAIN - UP TO 3X
			if ($i -lt 3) {
				Start-Sleep -Milliseconds 1000
				Continue
			} else {
				return "ERR/unable to get JSON 'clipstats' for '$AlertRec"
			}				
		}
		
		# Construct object to to returned
		if ($gif_method -match 'viaDAT') {
			$source_fs = $dat_file
		} else {
			$source_fs = $bvr_file
		}
		$rtn_obj = [ordered]@{}
		$rtn_obj.Add( 'method', $gif_method )
		$rtn_obj.Add( 'file', $output_fs )
		$rtn_obj.Add( 'delay', $delay_ms )
		$rtn_obj.Add( 'cnt', $gif_images_cnt )
		$rtn_obj.Add( 'height', $Height )
		$rtn_obj.Add( 'fps', $FPS )
		$rtn_obj.Add( 'kb', $gif_kb )
		$rtn_obj.Add( 'msec', $gif_dur_ms )
		$rtn_obj.Add( 'trimcust', $gif_dur_adj_pct )
		$rtn_obj.Add( 'aldur_ms', $al_dur_ms )
		#$rtn_obj.Add( 'res', $al_res )  # WxH
		$rtn_obj.Add( 'res', $gif_jpg_WxH )  # WxH
		$rtn_obj.Add( 'cam', $al_cam )
		$rtn_obj.Add( 'source', $source_fs )
		$rtn_obj.Add( 'archive', $archive_gif_fs )
		
		# For debugging
		#Write-Host ("rtn_obj cnt= " + ($rtn_obj.Count))
		#Write-Host ("rtn_obj = " + ($rtn_obj | ConvertTo-Json))
		
		return $rtn_obj
	}

	return "ERR/unable to make GIF"
}

function SendPushover() {
# PURPOSE: Send custom notification via the Pushover API
# ARGS:    $MsgStr          = (string) pushover notification message; may contain HTML
#          $MsgTitle        = (string) pushover notification title
#          $AttachPath      = (string) full filespec of resized alert image, or $null
#          $Camera          = (string) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
#          $PriorityLevel   = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
#          $FromConsoleFlg  = (boolean) $true notifies function that calling command was executed from the console
#          $QuietFlg        = (boolean) $true prevents background messaging (showing the message sent)
# RETURNS: (string) Pushover API response, or 'ERR/' message
# ATTRIBUTIONS:
#    handling attachments: https://tinyurl.com/mrxp7pn3  (PoshOver Powershell Package)
	
    param(
		[string]$MsgStr,            # [OPTIONAL] default is 'test'
		[string]$MsgTitle,          # [OPTIONAL] default is the App name associaled with your Pushover App token
		[int]$PriorityLevel,        # [OPTIONAL] default is 0
		[string]$AttachPath,        # [OPTIONAL] default is null
		[string]$Camera,            # [OPTIONAL] default is ""
		[boolean]$FromConsoleFlg,   # [OPTIONAL] default is $false
		[boolean]$QuietFlg          # [OPTIONAL] default is $false
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "$fn_name() MsgStr = '$MsgStr'"
	#Write-Host "$fn_name() MsgTitle = '$MsgTitle'"
	#Write-Host "$fn_name() PriorityLevel = PriorityLevel"
	#Write-Host "$fn_name() AttachPath = '$AttachPath'"	
	#Write-Host "$fn_name() QuietFlg = $QuietFlg"
	#Write-Host "$fn_name() fromConsole = FromConsoleFlg"
	# NOTE REGARDING MSG.EXE COMMANDS
	# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
	#msg.exe $script:server_username /TIME:20 "Pushover notification args...`npriority_level = $PriorityLevel`nQuietFlg = $QuietFlg`nAttachPath = $AttachPath`nMsgStr = $MsgStr"

	# HANDLE OPTIONAL COMMANDLINE ARGS
	if (!$MsgStr) {
		$MsgStr = "Alert w/ GIF attachment demo"
	}
	if (!$Camera) {
		$Camera = ""
	}

	$rtn_str = $null
	
	#Prep the attachment
	if ($AttachPath) {
		#Get-Item -Path $AttachPath
		$attach_obj = (Get-Item -Path $AttachPath)
		$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
	}

	# For debugging
	#Write-Host ("$fn_name() attach_obj = " + $attach_obj)
	#Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
	#Write-Host ("$fn_name() po_tags = " + $script:po_tags)
	#Write-Host ("$fn_name() Camera = " + $Camera)
	
	$uri = 'https://api.pushover.net/1/messages.json'
	$parameters = @{
		token      = $script:po_app_token
		user       = $script:po_user_token
		priority   = $PriorityLevel
		retry      = $script:po_retry
		expire     = $script:po_expire
		tags       = ($script:po_tags).Replace("&CAM", $Camera)
		sound      = $script:po_sound
		device     = $script:po_device
		html       = 1
		title      = $MsgTitle
		message    = $MsgStr
	}
	
	# For debugging
	#Write-Host ("parameters.Keys = " + $parameters.Keys)
	<# Write-Host "`nPushover parameters..."
	[string[]]$arr = $parameters | Out-String -stream
	foreach ($item in $arr) {
		if ($item) {
			Write-Host $item
		}
	} #>
	
	# EXECUTE
	if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {
		# SEND BASIC NOTIFICATION

		# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
		#	  do not execute; it doesn't produce anything, no error, warning, or output.
		#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
		#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
		#     all web requests now use basic parsing only.
		# ... attachmnets
		#     https://psg-prod-centralus.cloudapp.net/packages/Poshover/0.1.450600-alpha/Content/Private%5CSend-MessageWithAttachment.ps1
		$ApiResponse = $parameters | Invoke-RestMethod -Uri $uri -Method Post
		# Alternative syntaxes
		#$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null
		#$ApiResponse = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters

	} else {	
		# SEND NOTIFICATION WITH ATTACHMENT
		
		# Notes:
		# Borrowed code from:  https://github.com/joshooaj/Poshover/blob/main/src/Private/Send-MessageWithAttachment.ps1
		# More refs:           https://gist.github.com/markekraus/a0135d0a5f52b609eda513df4db71cab

		try {

            Add-Type -AssemblyName System.Net.Http
			# NOTE: Originally, the $client= line below was not working when this script was 
			#        executed in the background via a Blue Iris 'run a program' action
			#        SOLUTION:  https://stackoverflow.com/a/70535083
			$client = [system.net.http.httpclient]::new()

			try {
				$content = [system.net.http.multipartformdatacontent]::new()
				foreach ($key in $parameters.Keys) {
					$textContent = [system.net.http.stringcontent]::new($parameters.$key)
					$content.Add($textContent, $key)
				}
				
				# For debugging
				#Write-Host ("content = " + $content | ConvertTo-Json)
				
				$jpegContent = [system.net.http.bytearraycontent]::new($attach_bytes)
				$jpegContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
				$jpegContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
				$jpegContent.Headers.ContentDisposition.Name = 'attachment'
				$jpegContent.Headers.ContentDisposition.FileName = $AttachPath
				$content.Add($jpegContent)

				# Not needed.
				# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
				#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

				# Modify the message body
				Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.token, "********").Replace($parameters.user, "********"))"

				# For debugging
				#msg.exe $script:server_username /TIME:10 "executing pushover api cmd w/ attachment..."

				# DO IT!
				$ApiResponse = $client.PostAsync($uri, $content).Result
				
			}
			finally {
				$content.Dispose()  # Release resources
			}
		}
		finally {
			$client.Dispose()  # Release resources
		}
	}

	# PROVIDE FEEDBACK
	try {
		$response_str = $ApiResponse.Content.ReadAsStringAsync().Result | ConvertFrom-Json
	}
	catch {
		$response_str = "WARNING: Unexpected Pushover API response; attachment may be too large"
	}
	
	if ($ApiResponse.Content) {
		if ($response_str -match 'invalid') {
			$rtn_str = "ERR/Pushover API returned 'invalid'"
		} else {
			# SUCCESS!!
			$rtn_str = ("Pushover notification sent`n" + $response_str)
		}
	} else {
		$rtn_str = "ERR/Pushover API $ApiResponse.Content returned 'null'"
	}
	
	return $rtn_str
}

function LogPerformance() {
# PURPOSE:  Log performance data to CSV file for image file resizing and Pushover notification actions.
# ARGS:     see param lines below
# RETURNS:  (string) 'success' or 'ERR/unable to write to CSV log'
# NOTES:    1. CSV file is created automatically if it does not already exist. 
#           2. CSV file's header row is generated from the object keys in variable $new_row_obj.
# USAGE EXAMPLE:
# 	LogPerformance $csvfilespec $timestamp $images_cnt $gif_height $gif_fps $gif_trim_beg $gif_trim_end $gif_trim_pct_cust $gif_cur_ms $gif_size $makegif_ms $al_res $al_rec $al_file
	
    param(
		[string]$csvfilespec,   # [REQUIRED]
		[string]$timestamp,     # [REQUIRED]
		[string]$gif_source_fs,
		[string]$gif_method,
		[string]$gif_images, 
		[string]$gif_height, 
		[string]$gif_fps, 
		[string]$gif_trim_beg, 
		[string]$gif_trim_end, 
		[string]$gif_trim_pct_cust, 
		[string]$gif_dur_sec, 
		[string]$gif_size, 
		[double]$makegif_ms,
		[string]$gif_jpg_wxh,
		[string]$al_rec
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "$fn_name() timestamp = $timestamp"
	#Write-Host "$fn_name() al_file = $al_file"
	#Write-Host "$fn_name() al_rec = $al_rec"

	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$csvfilespec) {
		$err_msg = "ERR/expecting 2 arguments (csvfilespec timestamp)"
		return $err_msg
	}
	if (!$timestamp) {
		$err_msg = "ERR/expecting 2 arguments (csvfilespec timestamp)"
		return $err_msg
	}
	
	$rtn_str = 'success'
		
	# BUILD THE NEW ROW
	# ... Note: this object's keys are the header column labels
	$new_row_obj = [PSCustomObject]@{
		'Timestamp'      = $timestamp
		'SourceFile'     = $gif_source_fs
		'GifMethod'      = $gif_method
		'GifImagesCnt'   = $gif_images
		'GifHeight'      = $gif_height
		'GifFPS'         = $gif_fps
		'GifTrimBegPct'  = $gif_trim_beg
		'GifTrimEndPct'  = $gif_trim_end
		'GifTrimCustPct' = $gif_trim_end
		'GifDurSec'      = $gif_dur_sec
		'GifSizeKb'      = $gif_size
		'GifMakeMs'      = $makegif_ms
		'GifJpgWxH'      = $gif_jpg_wxh
		'AlertRecId'     = $al_rec
	}
	
	# For debugging
	#msg.exe $script:server_username /TIME:10 "new_row_obj... `n$new_row_obj"

	# APPEND NEW ROW TO CSV FILE, creating the file if necessary
	if ($PSVersionTable.PSVersion.Major -ge 7) {
		# REFS: https://tinyurl.com/yvy4mjy6
		#       https://stackoverflow.com/a/70717331
		#       https://stackoverflow.com/a/60681762
		try {
			$new_row_obj | Export-Csv -LiteralPath "$csvfilespec" -Append -Force -UseQuotes AsNeeded
		} catch {
			$rtn_str = "ERR/unable to write to CSV log"
		}
	} else {
		try {
			$new_row_obj | Export-Csv -LiteralPath "$csvfilespec" -Append -Force -NoTypeInformation
		} catch {
			$rtn_str = "ERR/unable to write to CSV log"
		}
	}
	
	return $rtn_str
}

function PostCustomBiLogEntry() {
# PURPOSE: Add custom message to the Blue Iris logfile
# ARGS:    $MsgStr        = (string) may contain HTML
#          $Level         = (integer) 0|1|2 (0:info, 1:warn, 2:error)
#          $QuietFlg     = (boolean) $true prevents background messaging
# RETURNS: (string) message as posted to the Blue Iris logfile

    param(
		[string]$MsgStr,      # [REQUIRED]
		[string]$Level,       # [OPTIONAL]
		[boolean]$QuietFlg   # [OPTIONAL]
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name()"
	#Write-Host "QuietFlg = $QuietFlg"

	# HANDLE REQUIRED COMMANDLINE ARGS
	if (!$MsgStr) {
		$err_msg = "ERR/expecting an argument (MsgStr)"
		return $err_msg
	}
	
	# Remove/replace incompatible substrings from $MsgStr
	$MsgStr = $MsgStr -Replace "`n", "%20" -Replace "\\r",  "%20" -Replace "\\n", "%20" -Replace " ", "%20"
	
	$bi_address = $script:bi_webserver
	if ($script:post_w_creds_flg) {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level&user=$script:bi_user_name&pw=$script:bi_user_pw"
	} else {
		$uri = "http://$bi_address/admin?log=$MsgStr&level=$Level"
	}
	$result = Invoke-WebRequest -Uri $uri
	
	# For debugging
	#Write-Host "uri = $uri"
	#Write-Host "result:`n$result"	

	# HANDLE NOTIFICATIONS
	if (!$QuietFlg) {
		#msg.exe $script:server_username /TIME:10 "Custom BI logfile message posted...`nlevel = $Level`n'$MsgStr'"
	}
	
	return $MsgStr
}

function PurgeOldFiles() {
# PURPOSE:  Delete old files (> N hours) from a filepath like 'D:\myfolder\*.gif'
# ARGS:     see param lines below
# RETURNS:  (array) @(found_files_cnt, purged_files_cnt)
#           or (string) 'ERR/' message
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
	
    param(
		[string]$Pattern,  # [REQUIRED] full filespec pattern for targeted files; like 'D:\myfolder\*.jpg'
		[int]$Hours        # [REQUIRED] maximum age for retained files; use 24 for 1 day, 168 for 1 week... 
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	if (!$Pattern -or !$Hours) {
		return "ERR/$fn_name() requires 2 arguments (-Pattern -Hours)"
	}

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	#Write-Host "$fn_name() Pattern = $Pattern"
	#Write-Host "$fn_name() Hours = $Hours"
	
	# GET FILES MATCHING PATTERN
	$cur_ary = get-childitem -Path $Pattern | Select Name
	if ($cur_ary) {
		$cur_ary_cnt = $cur_ary.Count
	} else {
		$cur_ary_cnt = 0
	}

	# For debugging
	#Write-Host "$fn_name() cur_ary cnt = $cur_ary_cnt"
	#Write-Host "$fn_name() cur_ary = $cur_ary"

	# GET FILES MATCHING AGE REQUIREMENT
	$purge_ary = @() 		
	if ($cur_ary_cnt -gt 0 -and $Hours) {
		
		$limit = (Get-Date).AddHours(-$Hours)
		
		# For debugging
		#Write-Host "$fn_name() limit = $limit"

		$purge_obj = Get-ChildItem -Path $Pattern | 
			Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
			Select Name

		if ($purge_obj) {
			$purge_ary = $purge_obj.Name
		}
	}
	
	# Get count for return array
	$purge_ary_cnt = $purge_ary.Count
	
	# For debugging
	#Write-Host "$fn_name() purge_ary cnt = $purge_ary"
	#Write-Host ("$fn_name() purge_ary = " + ($purge_ary | ConvertTo-Json))

	# REMOVE FILES MATCHING AGE REQUIREMENT
	if ($purge_ary_cnt -gt 0) {
		if ($Hours) {
			$null = Get-ChildItem -Path $Pattern | 
				Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
				Remove-Item -Force
		}
	}
	
	return  @($cur_ary_cnt, $purge_ary.Count)
}

function CopyFileAsSequence() {
# PURPOSE:  Copy a source file without over writing; append a sequence number '(N)' 
#           to the end of its basename until it is unique in the destination folder 
# ARGS:     see param lines below
# RETURNS:  (string) new_filespec, or 'fail'
# ATTRIBUTIONS:  https://stackoverflow.com/a/53595350/13499984

	param(
		[string]$SourceFilespec,   # [REQUIRED]
		[string]$TargetFilespec,  # [REQUIRED]
		[string]$DestPath         # [REQUIRED]
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	# For debugging
	#Write-Host "$fn_name() SourceFilespec = $SourceFilespec"
	#Write-Host "$fn_name() TargetFileName = $TargetFileName"
	#Write-Host "$fn_name() DestPath = $DestPath"
	
	$rtn_str = 'fail'  # Assume failure

	if ((Test-Path -LiteralPath "$SourceFilespec")) {

		if ((Test-Path -LiteralPath "$DestPath")) {

			$fileInfo = Get-Item -Path $TargetFilespec
			$base_name = $fileInfo.BaseName
			$extension = $fileInfo.Extension

			# Get an array of all filenames (name only) of the files with a similar name already present in the destination folder
			$allFiles = @(Get-ChildItem $DestPath -File -Filter "$base_name*$extension" | Select-Object -ExpandProperty Name)

			# Construct the new filename
			$new_name = $base_name + $extension  # Initialize
			$count = 1
			while ($allFiles -contains $new_name) {
				# add a sequence number in brackets to the end of the basename until it is unique in the destination folder
				$new_name = "{0}({1}){2}" -f $base_name, $count++, $extension
			}

			# Construct the new filespec
			$new_target_fs = Join-Path -Path $DestPath -ChildPath $new_name

			try {
				Copy-Item -Path $SourceFilespec -Destination $new_target_fs
				$rtn_str = $new_target_fs
			} catch {
				$rtn_str = "ERR/failed to create file '$new_target_fs'"
			}
		} else {
			$rtn_str = "ERR/destination folder does not exist ('$DestPath')"
		}
	} else {
		$rtn_str = "ERR/source file does not exist ('$FileName')"
	}
	
	return $rtn_str
		
}

function SleepWithStatusBar() {
# Source: https://gist.github.com/ctigeek/bd637eeaeeb71c5b17f4

	param (
		[double]$Seconds,
		[string]$Activity,
		[string]$Message
	)
	
    $doneDT = (Get-Date).AddSeconds($Seconds)
    while($doneDT -gt (Get-Date)) {
        $secondsLeft = $doneDT.Subtract((Get-Date)).TotalSeconds
        $percent = ($Seconds - $secondsLeft) / $Seconds * 100
 		Write-Progress -Activity $Activity -Status $Message -SecondsRemaining $secondsLeft -PercentComplete $percent
        [System.Threading.Thread]::Sleep(500)
    }
    Write-Progress -Activity $Activity -Status $Message -SecondsRemaining 0 -Completed
}

function PopupBackgroundMsg() {

	param(
		[string]$Message,   # [REQUIRED]
		[int]$Seconds,      # [REQUIRED]
		[string]$UserName,  # [REQUIRED]
		[boolean]$InBackground = $true
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	if ($InBackground) {
		msg.exe $UserName /TIME:$Seconds $Message
	}
}

# Start overall performance timer
$timer_total = [Diagnostics.Stopwatch]::StartNew()

if ($fromConsole -and !$Quiet) {
	if ($Help) {
		Write-Host "======================================================================================================"
		Write-Host "BLUE IRIS GIF TOOL - MAKE GIFS FROM CLIPS; SEND PUSHOVER NOTIFICATIONS WITH GIF ATTACHMENT"
		Write-Host "------------------------------------------------------------------------------------------------------"
		Write-Host "Usage: $ScriptName [args]"
		Write-Host "Args:  -AlertRec (or -AlertDB) [-AlertMemo -DelayPct]"
		Write-Host "        or -DbRec (or -UI3Rec) [-AlertMemo]"
		Write-Host "        or -Live -Camera [-DelayPct]"
		Write-Host "       -Seconds"
		Write-Host "       -ImagesCnt -Height -FPS"
		Write-Host "       -TrimBegPct -TrimEndPct"
		Write-Host "       -NukeDatFile"
		Write-Host "       -NoPush -NoBiLog -NoArchiveGIF"
		Write-Host "       -Quiet -Demo"
		Write-Host "       -Help"
		Write-Host "Note:  WHEN USING THIS SCRIPT FROM A BLUE IRIS ACTION SET..."
		Write-Host "       use [-AlertRec \&ALERT_DB] <-- KEY POINT: a backslash (or backtick '``') prefix is ESSENTIAL"
		Write-Host "       use [-AlertMemo &MEMO]"
		Write-Host "Note:  try [-Live] with [-DelayPct <100] to experiment with reducing the delay for BVR buffering"
		Write-Host "Note:  use [-IgnoreDatFile] to ignore a DAT file, which is used by default (if it exists)"
		Write-Host "Note:  use [-DbRec] or [-UI3Rec] to make GIFs from any existing alert with a duration"
		Write-Host "Note:  use [-Seconds] to change the default GIF duration; DEFAULT is 10 sec when using [-Live],"
		Write-Host "       else it's the Alert's duration (when using [-AlertRec] or [-UI3Rec]) "
		Write-Host "Note:  use [-TrimBegPct], [-TrimEndPct] make GIF from a portion of the [-AlertRec] clip"
		Write-Host "Note:  use user-setting '`$gif_trim_end_custom' to customize trim based on AI labels (e.g. 'vehicle')"
		Write-Host "Note:  use [-NukeDatFile] to remove DAT file from Blue Iris storage (always archived first)"
		Write-Host "Note:  use [-NoPush] to skip notification; also disables management of GIF size to below 2.5 MB"
		Write-Host "Note:  use [-Quiet] to suppress verbose output when executed from the console; shows CLI return object"
		Write-Host "Note:  use [-Demo] to skip notification/logfile/delete actions (useful when tuning GIF params)"
		Write-Host "Returns: (json string) if successful, or (string) if failure"
		Write-Host "    * (json) is like {'result: success, data: {gif_method: viaDat, gif_kb: 497.5, ...}}"
		Write-Host "    * (string) is like 'ERR/...'"
		Write-Host "------------------------------------------------------------------------------------------------------"
	}
}

# Initialize return object
$rtn_obj = @{}

# PARSE COMMANDLINE ARGS
# ... TEMPORARILY DISABLED
if ($NukeDatFile) {
	Write-Warning "Argument -NukeDatFile currently disabled"
	$NukeDatFile = $false
}
if (!$AlertRec -and !$AlertDB -and !$DbRec -and !$UI3Rec -and !$Live) {
	$err_msg = "ERR/expecting an argument (-AlertRec, -AlertDB, -DbRec, -UI3Rec, or -Live)"
	return $err_msg
}
if ($Live -and !$Camera) {
	$err_msg = "ERR/expecting an argument (-Camera) when using -Live"
	return $err_msg
}
if ($Live -and !$Camera) {
	$err_msg = "ERR/expecting an argument (-Camera) when using -Live"
	return $err_msg
}
$duration_sec = -1  # Use alert's duration via JSON cmd 'clipstats'
if ($Seconds) {
	$duration_sec = $Seconds
} else {
	if ($Live) {
		$duration_sec = 10
	}
}
if (!$DelayPct) {
	$DelayPct = 100
}
#Write-Host "duration_sec = $duration_sec"
if ($AlertRec) {
	$AlertRec = $AlertRec.Replace("\","")
}
if ($AlertDB -and !$AlertRec) {
	$AlertRec = $AlertRec.Replace("\","")
}
if ($UI3Rec -and !$DbRec) {
	$DbRec = $UI3Rec
}
#Write-Host "-Msg variable_test = $variable_test"
if ($Msg) {
	$po_msg = $Msg
} else {
	#$po_msg = "Alert w/ GIF attachment demo"
	$po_msg = ""
}
#Write-Host "po_msg = $po_msg"
if ($Priority -or ($Priority -eq 0)) {
	if ($Priority -ne '*') {
		if (($Priority -ge -2) -and ($Priority -le 2)) {
			$po_priority = $Priority
		}
	}
}
$gif_via_DAT_flg = $true  # Default
if($IgnoreDatFile) {
	$gif_via_DAT_flg = $false
}
#Write-Host "gif_via_DAT_flg = $gif_via_DAT_flg"
$nuke_DAT_file_flg = $false
if($NukeDatFile) {
	$nuke_DAT_file_flg = $true
} 
#Write-Host "nuke_DAT_file_flg = $nuke_DAT_file_flg"
if ($ImagesCnt) {
	$gif_images_cnt = $ImagesCnt
}
#Write-Host "gif_images_cnt = $gif_images_cnt"
if ($Height) {
	$gif_height = $Height
}
if ($FPS) {
	$gif_fps = $FPS
}
if ($TrimBegPct) {
	$gif_trim_beg = $TrimBegPct
}
if ($TrimEndPct) {
	$gif_trim_end = $TrimEndPct
}
#Write-Host "gif_trim_end = $gif_trim_end"
$nopush_flg = $false  # Default
if($NoPush) {
	$nopush_flg = $true
}
$nobilog_flg = $false  # Default
if ($NoBiLog) {
	$nobilog_flg = $true
}
$noarchivegif_flg = $false  # Default
if($NoArchiveGIF) {
	$noarchivegif_flg = $true
}
$quiet_flg = $false  # Default
if($Quiet) {
	$quiet_flg = $true
}
if ($quiet_flg) {
	$bg_msgs_on = $false
}
$demo_flg = $false  # Default
if($Demo) {
	$demo_flg = $true
}

# ABORT IF USER SETTINGS PATH NOT FOUND
if ($app_path) {
	if (-not(Test-Path -LiteralPath "$app_path") ) {
		$err_msg = "ERR/setting 'app_path' in user-settings file does not exist ('$app_path')`n"
		if ($post_to_BI_logfile_flg) {
			$null = PostCustomBiLogEntry `
				-MsgStr "$err_msg ($ScriptName)" `
				-Level 2 `
				-QuietFlg ($bg_msgs_on -eq $false)
		}
		if ($fromConsole) {
			return $err_msg
		} else {
			msg.exe $server_username /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
			exit
		}	
	}
}  else {
	$err_msg = "ERR/setting 'app_path' missing in user-settings file`n"
	if ($post_to_BI_logfile_flg) {
		$null = PostCustomBiLogEntry `
			-MsgStr "$err_msg ($ScriptName)" `
			-Level 2 `
			-QuietFlg ($bg_msgs_on -eq $false)
	}
	if ($fromConsole) {
		return $err_msg
	} else {
		msg.exe $server_username /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
		exit
	}	
}

if ($AlertRec -or $AlertDB -or $DbRec -or $UI3Rec -or $Live) {

	# Start resize performance timer
	$timer_makegif = [Diagnostics.Stopwatch]::StartNew()

	# For debugging
	#Write-Host "AlertRec = $AlertRec"
	#Write-Host "RootPath = $app_path"
	#Write-Host "UseDatFile = $gif_via_DAT_flg"
	#Write-Host "DatReqdImages = $gif_DAT_reqd_images"
	#Write-Host "NukeDatFile = $nuke_DAT_file_flg"
	#Write-Host "ImagesCnt = $gif_images_cnt"
	#Write-Host "TrimEndCust = $gif_trim_end_custom"
	#Write-Host "Height = $gif_height"
	#Write-Host "FPS = $gif_fps"
	#Write-Host "QuietFlg = $quiet_flg"

	# MAKE GIF ATTACHMENT
	$makegif_obj = `
		MakeGIF `
		-RootPath $app_path `
		-BvrPath $bi_new_folder_path `
		-AlertPath $bi_alert_folder_path `
		-AlertRec $AlertRec `
		-AlertMemo $AlertMemo `
		-Live $Live `
		-Camera $Camera `
		-DbRec $DbRec `
		-Seconds $duration_sec `
		-UseDatFile $gif_via_DAT_flg `
		-DatReqdImages $gif_DAT_reqd_images `
		-NoPushFlg $nopush_flg `
		-NukeDatFile $nuke_DAT_file_flg `
		-NoArchiveDatFlg $NoArchiveDat `
		-BackupDatFile $backup_DAT_file_flg `
		-NoArchiveGifFlg $noarchivegif_flg `
		-DelayPct $DelayPct `
		-ImagesCnt $gif_images_cnt `
		-ImagesCntUsedFlg $PSBoundParameters.ContainsKey('ImagesCnt') `
		-Height $gif_height `
		-FPS $gif_fps `
		-TrimBegPct $gif_trim_beg `
		-TrimEndPct $gif_trim_end `
		-TrimEndCust $gif_trim_end_custom `
		-QuietFlg $quiet_flg `
		-FromConsoleFlg $fromConsole
	
	# MakeGIF() RETURNS: 
	# {
	#	"method": "via Clip",
	#	"file": "D:\\pushover_gifs\\GT\\output.gif",
	#	"delay": 52612,
	#	"cnt": 10,
	#	"height": 320,
	#	"fps": 2,
	#	"msec": 2625,
	#	"kb": 226.2,
	#	"aldur_ms": 52612,
	#	"res": "1920x1080",
	#	"cam": "GT",
	#	"source": "GT.20230609_160000.bvr"
	#	"archive": "D:\pushover_gifs\_archive\GT.20230609_160000.971414.17-0.gif"
	# }

	# For debugging
	#Write-Host ("gif_response type = " + $makegif_obj.GetType())
	#Write-Host ("gif_response = " + ($makegif_obj | ConvertTo-Json))
	
	# Trap error conditions
	if ($makegif_obj.GetType() -match 'string') {
		if ($makegif_obj.StartsWith("ERR/") -or $makegif_obj.StartsWith("WARNING:")) {
			if ($FromConsoleFlg) {
				Write-Warning $makegif_obj 
			}
			return $makegif_obj
		}
	}
	
	# Parse output for items to be output and logged
	$gif_method    = $makegif_obj.method
	$gif_filespec  = $makegif_obj.file
	$delay_ms      = $makegif_obj.delay     # type is [int]
	$gif_imgs_cnt  = $makegif_obj.cnt       # type is [int]
	$gif_height    = $makegif_obj.height    # type is [int]
	$gif_fps       = $makegif_obj.fps       # type is [int]
	$gif_ms        = $makegif_obj.msec      # type is [int]
	$trim_cust_pct = $makegif_obj.trimcust  # type is [int]
	$gif_kb        = $makegif_obj.kb        # type is [double]
	$gif_jpg_wxh   = $makegif_obj.res
	$al_cam        = $makegif_obj.cam
	$gif_source_fs = $makegif_obj.source
	$archive_fs    = $makegif_obj.archive
	
	$gif_sec = [math]::Round($gif_ms/1000,1).ToString("#.0")
	$gif_source_fn = ($gif_source_fs.Replace(($bi_alert_folder_path),'')).TrimStart('\')
	
	# For debugging
	#Write-Host "gif_method = $gif_method"
	#Write-Host "delay_ms = $delay_ms"
	#Write-Host "gif_imgs_cnt = $gif_imgs_cnt"
	#Write-Host "gif_height = $gif_height"
	#Write-Host "gif_fps = $gif_fps"
	#Write-Host "gif_ms = $gif_ms"
	#Write-Host "gif_kb = $gif_kb"
	#Write-Host "gif_jpg_wxh = $gif_jpg_wxh"
	#Write-Host "al_cam = $al_cam"
	#Write-Host "gif_source_fs = $gif_source_fs"
	#Write-Host "gif archive_fs = $archive_fs"

	if ($fromConsole -and !$Quiet) {
		Write-Host "`nGIF File Summary:"
		Write-Host "-----------------"
		Write-Host "GIF created: '$gif_filespec' (via method: '$gif_method')"
		Write-Host "GIF specs: ImageCnt/Height/Dur(msec)/FPS/Size(KB) = $gif_imgs_cnt/$gif_height/$gif_ms/$gif_fps/$gif_kb"
	}
	if ($gif_kb -gt 2500) {
		Write-Host "WARNING: GIF file may exceed limit (2500 KB) for Pushover API"
	}
	if ($archive_fs -match 'fail') {
		Write-Host "WARNING: unable to archive the GIF file (reason unknown)"
	} elseif ($archive_fs.StartsWith("ERR/")) {
		Write-Host "WARNING: unable to archive the GIF file ($archive_fs)"
	} else {
		if ($fromConsole -and !$quiet_flg) {
			Write-Host "GIF archived as: '$archive_fs'"
		}
	}

	# Stop resize performance timer
	$timer_makegif.stop()
	#$makegif_ms = [math]::Round(($timer_makegif.elapsed.totalmilliseconds - $delay_ms),1).ToString("#.0")
	$makegif_ms = [math]::Round(($timer_makegif.elapsed.totalmilliseconds),1).ToString("#.0")
	
	# Fill return object
	#$rtn_obj = @{}
	$rtn_data_obj = [ordered]@{}
	$rtn_obj['result'] = 'success'
	$rtn_data_obj['gif_method'] = $gif_method
	$rtn_data_obj['gif_make_ms'] = $makegif_ms
	$rtn_data_obj['gif_images'] = $gif_imgs_cnt
	$rtn_data_obj['gif_height'] = $gif_height
	$rtn_data_obj['gif_fps'] = $gif_fps
	$rtn_data_obj['gif_ms'] = $gif_ms
	$rtn_data_obj['gif_trim_pct_beg'] = $TrimBegPct
	$rtn_data_obj['gif_trim_pct_end'] = $TrimEndPct
	$rtn_data_obj['gif_trim_pct_cust'] = $trim_cust_pct
	$rtn_data_obj['gif_kb'] = $gif_kb
	$rtn_data_obj['gif_jpg_wxh'] = $gif_jpg_wxh
	$rtn_data_obj['gif_fs'] = $archive_fs
	#$rtn_data_obj['gif_source_fs'] = $bi_new_folder_path + "\" + $gif_source_fs.Name
	$rtn_data_obj['gif_source_fs'] = $gif_source_fs
	$rtn_obj['data'] = $rtn_data_obj
}

# Start pushover performance timer
$timer_pusho = [Diagnostics.Stopwatch]::StartNew()

# SEND NOTIFICATION
if (!$demo_flg -and !$nopush_flg) {
	if ($Title) {
		$po_title = $Title
	} else {
		$po_title = "via $ScriptName"
	}
	
	
	if ($gif_kb -gt 0) {
		$gif_image_path = $app_path + $al_cam + "\output.gif"
		
		if (!$po_msg) {
			$po_msg = "GIF: imgs=$gif_imgs_cnt, H=$gif_height, sec=$gif_sec, FPS=$gif_fps, KB=$gif_kb, TrimBeg/End/Cust= $TrimBegPct/$TrimEndPct/$trim_cust_pct"
			if ($AlertMemo) {
				$po_msg = $po_msg + "; " + $AlertMemo
			} else {
				$po_msg = $po_msg + "; (no memo)"
			}
		}
		
		# For debugging
		#Write-Host "po_msg AFTER = '$po_msg'"

		# Append optional 'DAT tag' to Pushover message string
		if ($gif_method -match 'viaDAT') {
			if ($append_DAT_tag_flg) {
				$po_msg = ($po_msg + " (DAT" + $gif_imgs_cnt + ")")
			}
		}

		$po_response_str = `
			SendPushover `
			-MsgTitle $po_title `
			-MsgStr $po_msg `
			-AttachPath $gif_image_path `
			-Camera $al_cam `
			-PriorityLevel $po_priority `
			-FromConsoleFlg $fromConsole `
			-QuietFlg ($bg_msgs_on -eq $false)

		# PROVIDE FEEDBACK
		if ($bg_msgs_on) {
			$bg_msg = "Pushover notification sent...`npriority_level = $po_priority`n$MsgStr"
			#PopupBackgroundMsg -Seconds 15 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
			$bg_msg = "Pushover notification attachment...`n$gif_image_path"
			#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
			$bg_msg = "Pushover notification response...`n" + ($ApiResponse | ConvertTo-Json)
			#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
			$bg_msg = "Pushover notification response...`n" + $ApiResponse
			#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
		}

		if ($fromConsole -and !$Quiet) {
			Write-Host "`nPushover Summary:"
			Write-Host "-----------------"
		}
		if (($po_response_str.StartsWith("ERR/")) -or ($po_response_str.StartsWith("WARNING:"))) {
			if ($post_to_BI_logfile_flg) {
				$null = PostCustomBiLogEntry `
					-MsgStr "$po_response_str ($ScriptName)" `
					-Level 2 `
					-QuietFlg $quiet_flg
			}
			$bg_msg = "$po_response_str ($script:ScriptName)"
			PopupBackgroundMsg -Seconds 10 -Message $bg_msg -UserName $script:server_username -InBackground $bg_msgs_on
		}		
		if ($fromConsole -and !$Quiet) {
			Write-Host $po_response_str
		}
	} else {
		if ($fromConsole) {
			Write-Warning "Requested GIF attachment not created"
		}		
	}
}

# Stop pushover performance timer
$timer_pusho.stop()
$pusho_ms = [math]::Round($timer_pusho.elapsed.totalmilliseconds,1).ToString("#.0")

# Stop overall performance timer
$timer_total.stop()
$total_ms  = [math]::Round($timer_total.elapsed.totalmilliseconds,1).ToString("#.0")

# Log script performance info
# ... Prep values for logging
$csvfilespec = $app_path + $ScriptName + ".csv"
$timestamp   = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
$total_ms    = [decimal]$total_ms 
$pusho_ms    = [decimal]$pusho_ms 
#$gif_ms      = [decimal]$gif_ms 
#$gif_sec     = [decimal]$gif_sec

if ($gif_kb -gt 0) {
	$images_cnt        = [int]$gif_imgs_cnt
	$gif_height        = [int]$gif_height
	$gif_fps           = [int]$gif_fps
	$gif_trim_beg      = [int]$gif_trim_beg
	$gif_trim_end      = [int]$gif_trim_end
	$gif_trim_pct_cust = [int]$gif_trim_pct_cust
	$gif_sec           = [decimal]$gif_sec
	$gif_size          = [int]$gif_kb
	$makegif_ms        = [decimal]$makegif_ms
} else {
	$images_cnt        = 0
	$gif_height        = 0
	$gif_fps           = 0
	$gif_trim_beg      = 0
	$gif_trim_end      = 0
	$gif_trim_pct_cust = 0
	$gif_sec           = 0
	$gif_size          = 0
	$makegif_ms        = [decimal]$makegif_ms 
}

# ... Log the values
#     RETURNS:  (array) @(images_cnt, gif_height, gif_fps, gif_ms, gif_size, gif_filespec)
$log_response = `
	LogPerformance `
	$csvfilespec `
	$timestamp `
	$gif_source_fn `
	$gif_method `
	$images_cnt `
	$gif_height `
	$gif_fps `
	$gif_trim_beg `
	$gif_trim_end `
	$gif_trim_pct_cust `
	$gif_sec `
	$gif_size `
	$makegif_ms `
	$gif_jpg_wxh `
	$AlertRec 

if ($log_response.StartsWith("ERR/")) {
	Write-Host $log_response
	if (!$fromConsole) {
		$bg_msg = $log_response
		PopupBackgroundMsg -Seconds 10 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
	}
}

$disp_total_ms   = [math]::Round($total_ms,0)
$disp_delay_ms   = [math]::Round($delay_ms,0)
$disp_makegif_ms = [math]::Round($makegif_ms,0)
$disp_pusho_ms   = [math]::Round($pusho_ms,0)

# Performance notice
if ($fromConsole -and !$quiet_flg) {
	$perf_msg = (`
		"Performance Summary:" `
		+ "`n--------------------" `
		+ "`nJPG fetch delay:         $disp_delay_ms msec" `
		+ "`nmakegif execution time:  $disp_makegif_ms msec" `
		+ "`npushover execution time: $disp_pusho_ms msec" `
		+ "`ntotal execution time:    $disp_total_ms msec"`
		)
	Write-Host "`n$perf_msg"
}

#Write-Host "bg_msgs_on = $bg_msgs_on"

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
if ($bg_msgs_on) {
	$bg_msg =  "Alert record: $AlertRec" + 
		"`nclipstats file: $al_file" +
		"`n`nGIF Height/FPS/Dur(s)/Size(KB): $gif_height / $gif_fps / $gif_sec / $gif_kb" + 
		"`n`nTimes (delay,makegif,pusho,total) ... $disp_delay_ms, $disp_makegif_ms, $disp_pusho_ms, $disp_total_ms msec" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"
		
	# KEY POINT: message length limit is 255 characters
	# For debugging.
	#Write-Host ("bg_msg len = " + $bg_msg.length)
	#Write-Host ("bg_msg:`n" + $bg_msg)
	
	PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
}

# POST MESSAGE TO BLUE IRIS LOGFILE
$bilogfile_msg = "No custom BI logfile message posted"
if (!$nobilog_flg) {
	if ($post_all_events) {
		if (!$demo_flg) {
			if ($rtn_obj -match "ERR/") {
				$log_level = 2
				$log_msg = "$scriptName - $rtn_obj"
			} elseif ($rtn_obj -match "WARNING:") {
				$log_level = 1
				$log_msg = "$scriptName - $rtn_obj"
			} else {
				$log_level = 0
				$log_msg = "$ScriptBaseName [$gif_height H, $gif_fps FPS, $gif_sec sec, $gif_kb KB] [$disp_delay_ms ms, $disp_makegif_ms ms, $disp_pusho_ms ms, $disp_total_ms ms] method=$gif_method rec=$AlertRec source=$gif_source_fn"
			}
			$response_str = `
				PostCustomBiLogEntry `
				-MsgStr "$log_msg" `
				-Level $log_level `
				-QuietFlg $true
				
			if ($response_str) {
				$bilogfile_msg = "Custom BI logfile message posted"
			}
		}
	}
}
if ($fromConsole -and !$Quiet) {
	Write-Host "`nBlue Iris Logfile:"
	Write-Host "------------------"
	Write-Host $bilogfile_msg
}

if ($fromConsole -and $Quiet) {
	#Write-Host "return object ="
}

if (!$fromConsole -or $Quiet) {
	if ($RtnHashtable) {
		return $rtn_obj
	} else {
		return ($rtn_obj | ConvertTo-Json)
	}
}

if ($fromConsole -and !$Quiet) {
	Write-Host ""
}

#Stop-Transcript